from __future__ import annotations

from typing import TYPE_CHECKING

from FL.helpers.registry import parse_registry_file

if TYPE_CHECKING:
    from pathlib import Path


class Options:
    """
    Options() - generic constructor
    """

    # Constructor

    def __init__(self) -> None:
        """
        # No args
        >>> op = Options()
        >>> print(op)
        <Options>
        """
        self.Init()

    def __repr__(self) -> str:
        return "<Options>"

    # Additions for FakeLab

    def fake_load_regfile(self, file_path: Path) -> None:
        """
        Load options from a registry file
        """
        # TODO
        print(f"Loading options from {file_path.name}...")
        reg = parse_registry_file(file_path)
        for k, v in reg.items():
            if hasattr(self, k):
                setattr(self, k, v)
            else:
                print(f"Unsupported option: '{k}'")
        print("...done.")

    def fake_save_regfile(self, file_path: Path) -> None:
        """
        Save options to a registry file
        """
        # TODO
        header = (
            "Windows Registry Editor Version 5.00\n\n"
            "[HKEY_CURRENT_USER\\Software\\FontLab\\FontStudio 5\\Options]\n"
        )
        print(f"Save options {header} to {file_path}")
        raise NotImplementedError

    # Attributes

    @property
    def AutoAlignVector(self) -> int:
        return self._AutoAlignVector

    @AutoAlignVector.setter
    def AutoAlignVector(self, value: int) -> None:
        self._AutoAlignVector = value

    @property
    def AutoMetricsClose(self) -> int:
        return self._AutoMetricsClose

    @AutoMetricsClose.setter
    def AutoMetricsClose(self, value: int) -> None:
        self._AutoMetricsClose = value

    @property
    def AutoMetricsLeft(self) -> int:
        return self._AutoMetricsLeft

    @AutoMetricsLeft.setter
    def AutoMetricsLeft(self, value: int) -> None:
        self._AutoMetricsLeft = value

    @property
    def AutoMetricsRight(self) -> int:
        return self._AutoMetricsRight

    @AutoMetricsRight.setter
    def AutoMetricsRight(self, value: int) -> None:
        self._AutoMetricsRight = value

    @property
    def AutoRemoveHints(self) -> int:
        return self._AutoRemoveHints

    @AutoRemoveHints.setter
    def AutoRemoveHints(self, value: int) -> None:
        self._AutoRemoveHints = value

    @property
    def AutoSave(self) -> int:
        return self._AutoSave

    @AutoSave.setter
    def AutoSave(self, value: int) -> None:
        self._AutoSave = value

    @property
    def AutoSaveTick(self) -> int:
        return self._AutoSaveTick

    @AutoSaveTick.setter
    def AutoSaveTick(self, value: int) -> None:
        self._AutoSaveTick = value

    @property
    def AutohintingHRatio(self) -> float:
        return self._AutohintingHRatio

    @AutohintingHRatio.setter
    def AutohintingHRatio(self, value: float) -> None:
        self._AutohintingHRatio = value

    @property
    def AutohintingMaxHWidth(self) -> int:
        return self._AutohintingMaxHWidth

    @AutohintingMaxHWidth.setter
    def AutohintingMaxHWidth(self, value: int) -> None:
        self._AutohintingMaxHWidth = value

    @property
    def AutohintingMaxVWidth(self) -> int:
        return self._AutohintingMaxVWidth

    @AutohintingMaxVWidth.setter
    def AutohintingMaxVWidth(self, value: int) -> None:
        self._AutohintingMaxVWidth = value

    @property
    def AutohintingMinHLen(self) -> int:
        return self._AutohintingMinHLen

    @AutohintingMinHLen.setter
    def AutohintingMinHLen(self, value: int) -> None:
        self._AutohintingMinHLen = value

    @property
    def AutohintingMinHWidth(self) -> int:
        return self._AutohintingMinHWidth

    @AutohintingMinHWidth.setter
    def AutohintingMinHWidth(self, value: int) -> None:
        self._AutohintingMinHWidth = value

    @property
    def AutohintingMinVLen(self) -> int:
        return self._AutohintingMinVLen

    @AutohintingMinVLen.setter
    def AutohintingMinVLen(self, value: int) -> None:
        self._AutohintingMinVLen = value

    @property
    def AutohintingMinVWidth(self) -> int:
        return self._AutohintingMinVWidth

    @AutohintingMinVWidth.setter
    def AutohintingMinVWidth(self, value: int) -> None:
        self._AutohintingMinVWidth = value

    @property
    def AutohintingRemoveHints(self) -> int:
        return self._AutohintingRemoveHints

    @AutohintingRemoveHints.setter
    def AutohintingRemoveHints(self, value: int) -> None:
        self._AutohintingRemoveHints = value

    @property
    def AutohintingVRatio(self) -> float:
        return self._AutohintingVRatio

    @AutohintingVRatio.setter
    def AutohintingVRatio(self, value: float) -> None:
        self._AutohintingVRatio = value

    @property
    def Backup(self) -> int:
        return self._Backup

    @Backup.setter
    def Backup(self, value: int) -> None:
        self._Backup = value

    @property
    def BitmapSize(self) -> int:
        return self._BitmapSize

    @BitmapSize.setter
    def BitmapSize(self, value: int) -> None:
        self._BitmapSize = value

    @property
    def ChartApplyTemplate(self) -> int:
        return self._ChartApplyTemplate

    @ChartApplyTemplate.setter
    def ChartApplyTemplate(self, value: int) -> None:
        self._ChartApplyTemplate = value

    @property
    def ChartAutoHide(self) -> int:
        return self._ChartAutoHide

    @ChartAutoHide.setter
    def ChartAutoHide(self, value: int) -> None:
        self._ChartAutoHide = value

    @property
    def ChartAutoUnicode(self) -> int:
        return self._ChartAutoUnicode

    @ChartAutoUnicode.setter
    def ChartAutoUnicode(self, value: int) -> None:
        self._ChartAutoUnicode = value

    @property
    def ChartCaptionFontCharSet(self) -> int:
        return self._ChartCaptionFontCharSet

    @ChartCaptionFontCharSet.setter
    def ChartCaptionFontCharSet(self, value: int) -> None:
        self._ChartCaptionFontCharSet = value

    @property
    def ChartCaptionFontName(self) -> str:
        return self._ChartCaptionFontName

    @ChartCaptionFontName.setter
    def ChartCaptionFontName(self, value: str) -> None:
        self._ChartCaptionFontName = value

    @property
    def ChartCaptionFontSize(self) -> int:
        return self._ChartCaptionFontSize

    @ChartCaptionFontSize.setter
    def ChartCaptionFontSize(self, value: int) -> None:
        self._ChartCaptionFontSize = value

    @property
    def ChartDoubleClick(self) -> int:
        return self._ChartDoubleClick

    @ChartDoubleClick.setter
    def ChartDoubleClick(self, value: int) -> None:
        self._ChartDoubleClick = value

    @property
    def ChartDragDrop(self) -> int:
        return self._ChartDragDrop

    @ChartDragDrop.setter
    def ChartDragDrop(self, value: int) -> None:
        self._ChartDragDrop = value

    @property
    def ChartPreviewMarks(self) -> int:
        return self._ChartPreviewMarks

    @ChartPreviewMarks.setter
    def ChartPreviewMarks(self, value: int) -> None:
        self._ChartPreviewMarks = value

    @property
    def ChartSampleSize(self) -> int:
        return self._ChartSampleSize

    @ChartSampleSize.setter
    def ChartSampleSize(self, value: int) -> None:
        self._ChartSampleSize = value

    @property
    def ChartShowNotes(self) -> int:
        return self._ChartShowNotes

    @ChartShowNotes.setter
    def ChartShowNotes(self, value: int) -> None:
        self._ChartShowNotes = value

    @property
    def ChartSorting(self) -> int:
        return self._ChartSorting

    @ChartSorting.setter
    def ChartSorting(self, value: int) -> None:
        self._ChartSorting = value

    @property
    def ChartStyle(self) -> int:
        return self._ChartStyle

    @ChartStyle.setter
    def ChartStyle(self, value: int) -> None:
        self._ChartStyle = value

    @property
    def ChartTemplateFontCharSet(self) -> int:
        return self._ChartTemplateFontCharSet

    @ChartTemplateFontCharSet.setter
    def ChartTemplateFontCharSet(self, value: int) -> None:
        self._ChartTemplateFontCharSet = value

    @property
    def ChartTemplateFontName(self) -> str:
        return self._ChartTemplateFontName

    @ChartTemplateFontName.setter
    def ChartTemplateFontName(self, value: str) -> None:
        self._ChartTemplateFontName = value

    @property
    def ChartTemplateFontSize(self) -> int:
        return self._ChartTemplateFontSize

    @ChartTemplateFontSize.setter
    def ChartTemplateFontSize(self, value: int) -> None:
        self._ChartTemplateFontSize = value

    @property
    def ChartUseTemplate(self) -> int:
        return self._ChartUseTemplate

    @ChartUseTemplate.setter
    def ChartUseTemplate(self, value: int) -> None:
        self._ChartUseTemplate = value

    @property
    def ClosepathArrowLen(self) -> int:
        return self._ClosepathArrowLen

    @ClosepathArrowLen.setter
    def ClosepathArrowLen(self, value: int) -> None:
        self._ClosepathArrowLen = value

    @property
    def CodepagesCount(self) -> int:
        return self._CodepagesCount

    @CodepagesCount.setter
    def CodepagesCount(self, value: int) -> None:
        self._CodepagesCount = value

    @property
    def ColorBitmap(self) -> int:
        return self._ColorBitmap

    @ColorBitmap.setter
    def ColorBitmap(self, value: int) -> None:
        self._ColorBitmap = value

    @property
    def ColorClosepath(self) -> int:
        return self._ColorClosepath

    @ColorClosepath.setter
    def ColorClosepath(self, value: int) -> None:
        self._ColorClosepath = value

    @property
    def ColorEcho(self) -> int:
        return self._ColorEcho

    @ColorEcho.setter
    def ColorEcho(self, value: int) -> None:
        self._ColorEcho = value

    @property
    def ColorHintsPen(self) -> int:
        return self._ColorHintsPen

    @ColorHintsPen.setter
    def ColorHintsPen(self, value: int) -> None:
        self._ColorHintsPen = value

    @property
    def ColorMaskPen(self) -> int:
        return self._ColorMaskPen

    @ColorMaskPen.setter
    def ColorMaskPen(self, value: int) -> None:
        self._ColorMaskPen = value

    @property
    def ColorSeacPen(self) -> int:
        return self._ColorSeacPen

    @ColorSeacPen.setter
    def ColorSeacPen(self, value: int) -> None:
        self._ColorSeacPen = value

    @property
    def ColorTemplate(self) -> int:
        return self._ColorTemplate

    @ColorTemplate.setter
    def ColorTemplate(self, value: int) -> None:
        self._ColorTemplate = value

    @property
    def ColorVMetrics(self) -> int:
        return self._ColorVMetrics

    @ColorVMetrics.setter
    def ColorVMetrics(self, value: int) -> None:
        self._ColorVMetrics = value

    @property
    def CopyHDMXData(self) -> int:
        """
        Copy HDMX data from base to composite glyph?
        """
        return self._CopyHDMXData

    @CopyHDMXData.setter
    def CopyHDMXData(self, value: int) -> None:
        self._CopyHDMXData = value

    @property
    def CreateUnexistingCharacters(self) -> int:
        return self._CreateUnexistingCharacters

    @CreateUnexistingCharacters.setter
    def CreateUnexistingCharacters(self, value: int) -> None:
        self._CreateUnexistingCharacters = value

    @property
    def DefaultGlyph(self) -> int:
        return self._DefaultGlyph

    @DefaultGlyph.setter
    def DefaultGlyph(self, value: int) -> None:
        self._DefaultGlyph = value

    @property
    def DuplicateX(self) -> int:
        return self._DuplicateX

    @DuplicateX.setter
    def DuplicateX(self, value: int) -> None:
        self._DuplicateX = value

    @property
    def DuplicateY(self) -> int:
        return self._DuplicateY

    @DuplicateY.setter
    def DuplicateY(self, value: int) -> None:
        self._DuplicateY = value

    @property
    def EditBCPsFixed(self) -> int:
        return self._EditBCPsFixed

    @EditBCPsFixed.setter
    def EditBCPsFixed(self, value: int) -> None:
        self._EditBCPsFixed = value

    @property
    def EditBitmapStyle(self) -> int:
        return self._EditBitmapStyle

    @EditBitmapStyle.setter
    def EditBitmapStyle(self, value: int) -> None:
        self._EditBitmapStyle = value

    @property
    def EditChangeCursor(self) -> int:
        return self._EditChangeCursor

    @EditChangeCursor.setter
    def EditChangeCursor(self, value: int) -> None:
        self._EditChangeCursor = value

    @property
    def EditFollowScroll(self) -> int:
        return self._EditFollowScroll

    @EditFollowScroll.setter
    def EditFollowScroll(self, value: int) -> None:
        self._EditFollowScroll = value

    @property
    def EditGridX(self) -> int:
        return self._EditGridX

    @EditGridX.setter
    def EditGridX(self, value: int) -> None:
        self._EditGridX = value

    @property
    def EditGridY(self) -> int:
        return self._EditGridY

    @EditGridY.setter
    def EditGridY(self, value: int) -> None:
        self._EditGridY = value

    @property
    def EditHandleTool(self) -> int:
        return self._EditHandleTool

    @EditHandleTool.setter
    def EditHandleTool(self, value: int) -> None:
        self._EditHandleTool = value

    @property
    def EditHitDistance(self) -> int:
        return self._EditHitDistance

    @EditHitDistance.setter
    def EditHitDistance(self, value: int) -> None:
        self._EditHitDistance = value

    @property
    def EditInstantRefresh(self) -> int:
        return self._EditInstantRefresh

    @EditInstantRefresh.setter
    def EditInstantRefresh(self, value: int) -> None:
        self._EditInstantRefresh = value

    @property
    def EditLeaveEcho(self) -> int:
        return self._EditLeaveEcho

    @EditLeaveEcho.setter
    def EditLeaveEcho(self, value: int) -> None:
        self._EditLeaveEcho = value

    @property
    def EditNotFillOpen(self) -> int:
        return self._EditNotFillOpen

    @EditNotFillOpen.setter
    def EditNotFillOpen(self, value: int) -> None:
        self._EditNotFillOpen = value

    @property
    def EditRulers(self) -> int:
        return self._EditRulers

    @EditRulers.setter
    def EditRulers(self, value: int) -> None:
        self._EditRulers = value

    @property
    def EditScaleEPS(self) -> int:
        return self._EditScaleEPS

    @EditScaleEPS.setter
    def EditScaleEPS(self, value: int) -> None:
        self._EditScaleEPS = value

    @property
    def EditShowCross(self) -> int:
        return self._EditShowCross

    @EditShowCross.setter
    def EditShowCross(self, value: int) -> None:
        self._EditShowCross = value

    @property
    def EditShowPosition(self) -> int:
        return self._EditShowPosition

    @EditShowPosition.setter
    def EditShowPosition(self, value: int) -> None:
        self._EditShowPosition = value

    @property
    def EditSmallNodes(self) -> int:
        return self._EditSmallNodes

    @EditSmallNodes.setter
    def EditSmallNodes(self, value: int) -> None:
        self._EditSmallNodes = value

    @property
    def EditSmoothOutline(self) -> int:
        return self._EditSmoothOutline

    @EditSmoothOutline.setter
    def EditSmoothOutline(self, value: int) -> None:
        self._EditSmoothOutline = value

    @property
    def EraseSize(self) -> int:
        return self._EraseSize

    @EraseSize.setter
    def EraseSize(self, value: int) -> None:
        self._EraseSize = value

    @property
    def FitAscender(self) -> int:
        return self._FitAscender

    @FitAscender.setter
    def FitAscender(self, value: int) -> None:
        self._FitAscender = value

    @property
    def FitDescender(self) -> int:
        return self._FitDescender

    @FitDescender.setter
    def FitDescender(self, value: int) -> None:
        self._FitDescender = value

    @property
    def FontAudit(self) -> int:
        return self._FontAudit

    @FontAudit.setter
    def FontAudit(self, value: int) -> None:
        self._FontAudit = value

    @property
    def GlyphsBarOpen(self) -> int:
        return self._GlyphsBarOpen

    @GlyphsBarOpen.setter
    def GlyphsBarOpen(self, value: int) -> None:
        self._GlyphsBarOpen = value

    @property
    def HideAllLayers(self) -> int:
        return self._HideAllLayers

    @HideAllLayers.setter
    def HideAllLayers(self, value: int) -> None:
        self._HideAllLayers = value

    @property
    def HideToolbars(self) -> int:
        return self._HideToolbars

    @HideToolbars.setter
    def HideToolbars(self, value: int) -> None:
        self._HideToolbars = value

    @property
    def LockStyleEx(self) -> int:
        return self._LockStyleEx

    @LockStyleEx.setter
    def LockStyleEx(self, value: int) -> None:
        self._LockStyleEx = value

    @property
    def OTAddClasses(self) -> int:
        return self._OTAddClasses

    @OTAddClasses.setter
    def OTAddClasses(self, value: int) -> None:
        self._OTAddClasses = value

    @property
    def OTWriteGDEF(self) -> int:
        return self._OTWriteGDEF

    @OTWriteGDEF.setter
    def OTWriteGDEF(self, value: int) -> None:
        self._OTWriteGDEF = value

    @property
    def OptimizeAlign(self) -> int:
        return self._OptimizeAlign

    @OptimizeAlign.setter
    def OptimizeAlign(self, value: int) -> None:
        self._OptimizeAlign = value

    @property
    def OptimizeReduce(self) -> int:
        return self._OptimizeReduce

    @OptimizeReduce.setter
    def OptimizeReduce(self, value: int) -> None:
        self._OptimizeReduce = value

    @property
    def OverlapMode(self) -> int:
        return self._OverlapMode

    @OverlapMode.setter
    def OverlapMode(self, value: int) -> None:
        self._OverlapMode = value

    @property
    def PaintAutoTransform(self) -> int:
        return self._PaintAutoTransform

    @PaintAutoTransform.setter
    def PaintAutoTransform(self, value: int) -> None:
        self._PaintAutoTransform = value

    @property
    def PaintAutoView(self) -> int:
        return self._PaintAutoView

    @PaintAutoView.setter
    def PaintAutoView(self, value: int) -> None:
        self._PaintAutoView = value

    @property
    def PaintBrushBody(self) -> int:
        return self._PaintBrushBody

    @PaintBrushBody.setter
    def PaintBrushBody(self, value: int) -> None:
        self._PaintBrushBody = value

    @property
    def PaintBrushCap(self) -> int:
        return self._PaintBrushCap

    @PaintBrushCap.setter
    def PaintBrushCap(self, value: int) -> None:
        self._PaintBrushCap = value

    @property
    def PaintBrushJoin(self) -> int:
        return self._PaintBrushJoin

    @PaintBrushJoin.setter
    def PaintBrushJoin(self, value: int) -> None:
        self._PaintBrushJoin = value

    @property
    def PaintBrushRoundness(self) -> int:
        return self._PaintBrushRoundness

    @PaintBrushRoundness.setter
    def PaintBrushRoundness(self, value: int) -> None:
        self._PaintBrushRoundness = value

    @property
    def PaintBrushVectorX(self) -> float:
        return self._PaintBrushVectorX

    @PaintBrushVectorX.setter
    def PaintBrushVectorX(self, value: float) -> None:
        self._PaintBrushVectorX = value

    @property
    def PaintBrushVectorY(self) -> float:
        return self._PaintBrushVectorY

    @PaintBrushVectorY.setter
    def PaintBrushVectorY(self, value: float) -> None:
        self._PaintBrushVectorY = value

    @property
    def PaintBrushWidth(self) -> int:
        return self._PaintBrushWidth

    @PaintBrushWidth.setter
    def PaintBrushWidth(self, value: int) -> None:
        self._PaintBrushWidth = value

    @property
    def PaintColor(self) -> int:
        return self._PaintColor

    @PaintColor.setter
    def PaintColor(self, value: int) -> None:
        self._PaintColor = value

    @property
    def PaintMode(self) -> int:
        return self._PaintMode

    @PaintMode.setter
    def PaintMode(self, value: int) -> None:
        self._PaintMode = value

    @property
    def PaintStyleEx(self) -> int:
        return self._PaintStyleEx

    @PaintStyleEx.setter
    def PaintStyleEx(self, value: int) -> None:
        self._PaintStyleEx = value

    @property
    def PaintTextSize(self) -> int:
        return self._PaintTextSize

    @PaintTextSize.setter
    def PaintTextSize(self, value: int) -> None:
        self._PaintTextSize = value

    @property
    def PaintTool(self) -> int:
        return self._PaintTool

    @PaintTool.setter
    def PaintTool(self, value: int) -> None:
        self._PaintTool = value

    @property
    def PaintTracePolygon(self) -> int:
        return self._PaintTracePolygon

    @PaintTracePolygon.setter
    def PaintTracePolygon(self, value: int) -> None:
        self._PaintTracePolygon = value

    @property
    def PasteX(self) -> int:
        return self._PasteX

    @PasteX.setter
    def PasteX(self, value: int) -> None:
        self._PasteX = value

    @property
    def PasteY(self) -> int:
        return self._PasteY

    @PasteY.setter
    def PasteY(self, value: int) -> None:
        self._PasteY = value

    @property
    def PreviewExpanded(self) -> int:
        return self._PreviewExpanded

    @PreviewExpanded.setter
    def PreviewExpanded(self, value: int) -> None:
        self._PreviewExpanded = value

    @property
    def PreviewExpandedHeight(self) -> int:
        return self._PreviewExpandedHeight

    @PreviewExpandedHeight.setter
    def PreviewExpandedHeight(self, value: int) -> None:
        self._PreviewExpandedHeight = value

    @property
    def PreviewPPMs(self) -> str:
        return self._PreviewPPMs

    @PreviewPPMs.setter
    def PreviewPPMs(self, value: str) -> None:
        self._PreviewPPMs = value

    @property
    def PreviewPPMsExpanded(self) -> str:
        return self._PreviewPPMsExpanded

    @PreviewPPMsExpanded.setter
    def PreviewPPMsExpanded(self, value: str) -> None:
        self._PreviewPPMsExpanded = value

    @property
    def PreviewPointSize(self) -> str:
        return self._PreviewPointSize

    @PreviewPointSize.setter
    def PreviewPointSize(self, value: str) -> None:
        self._PreviewPointSize = value

    @property
    def PreviewSecondLine(self) -> str:
        return self._PreviewSecondLine

    @PreviewSecondLine.setter
    def PreviewSecondLine(self, value: str) -> None:
        self._PreviewSecondLine = value

    @property
    def PreviewSmooth(self) -> int:
        return self._PreviewSmooth

    @PreviewSmooth.setter
    def PreviewSmooth(self, value: int) -> None:
        self._PreviewSmooth = value

    @property
    def PreviewWidth(self) -> int:
        return self._PreviewWidth

    @PreviewWidth.setter
    def PreviewWidth(self, value: int) -> None:
        self._PreviewWidth = value

    @property
    def SamplePPM1(self) -> int:
        return self._SamplePPM1

    @SamplePPM1.setter
    def SamplePPM1(self, value: int) -> None:
        self._SamplePPM1 = value

    @property
    def SamplePPM2(self) -> int:
        return self._SamplePPM2

    @SamplePPM2.setter
    def SamplePPM2(self, value: int) -> None:
        self._SamplePPM2 = value

    @property
    def SnapStyleEx(self) -> int:
        return self._SnapStyleEx

    @SnapStyleEx.setter
    def SnapStyleEx(self, value: int) -> None:
        self._SnapStyleEx = value

    @property
    def T1AFM(self) -> int:
        return self._T1AFM

    @T1AFM.setter
    def T1AFM(self, value: int) -> None:
        self._T1AFM = value

    @property
    def T1Autohint(self) -> int:
        return self._T1Autohint

    @T1Autohint.setter
    def T1Autohint(self, value: int) -> None:
        self._T1Autohint = value

    @property
    def T1Decompose(self) -> int:
        return self._T1Decompose

    @T1Decompose.setter
    def T1Decompose(self, value: int) -> None:
        self._T1Decompose = value

    @property
    def T1Encoding(self) -> int:
        return self._T1Encoding

    @T1Encoding.setter
    def T1Encoding(self, value: int) -> None:
        self._T1Encoding = value

    @property
    def T1PFM(self) -> int:
        return self._T1PFM

    @T1PFM.setter
    def T1PFM(self, value: int) -> None:
        self._T1PFM = value

    @property
    def T1Terminal(self) -> int:
        return self._T1Terminal

    @T1Terminal.setter
    def T1Terminal(self, value: int) -> None:
        self._T1Terminal = value

    @property
    def T1Unicode(self) -> int:
        return self._T1Unicode

    @T1Unicode.setter
    def T1Unicode(self, value: int) -> None:
        self._T1Unicode = value

    @property
    def T1UseOS2(self) -> int:
        return self._T1UseOS2

    @T1UseOS2.setter
    def T1UseOS2(self, value: int) -> None:
        self._T1UseOS2 = value

    @property
    def TTEAddCharacters(self) -> int:
        return self._TTEAddCharacters

    @TTEAddCharacters.setter
    def TTEAddCharacters(self, value: int) -> None:
        self._TTEAddCharacters = value

    @property
    def TTEApplyBBoxSavings(self) -> int:
        return self._TTEApplyBBoxSavings

    @TTEApplyBBoxSavings.setter
    def TTEApplyBBoxSavings(self, value: int) -> None:
        self._TTEApplyBBoxSavings = value

    @property
    def TTEAutoWinAscDesc(self) -> int:
        return self._TTEAutoWinAscDesc

    @TTEAutoWinAscDesc.setter
    def TTEAutoWinAscDesc(self, value: int) -> None:
        self._TTEAutoWinAscDesc = value

    @property
    def TTEAutohint(self) -> int:
        """
        Autohint unhinted glyphs
        """
        return self._TTEAutohint

    @TTEAutohint.setter
    def TTEAutohint(self, value: int) -> None:
        self._TTEAutohint = value

    @property
    def TTECmap10(self) -> int:
        """
        option:   - Use following codepage to build cmap(1,0) table:
                    [Current codepage in the Font Window]

        Returns:
            int: _description_
        """
        return self._TTECmap10

    @TTECmap10.setter
    def TTECmap10(self, value: int) -> None:
        self._TTECmap10 = value

    @property
    def TTEExportOT(self) -> int:
        """
        Export OpenType layout tables?
        """
        return self._TTEExportOT

    @TTEExportOT.setter
    def TTEExportOT(self, value: int) -> None:
        self._TTEExportOT = value

    @property
    def TTEExportUnicode(self) -> int:
        """_summary_

        Returns:
            int: _description_

        checked:  - Ignore Unicode indexes in the font
        option    - Use following codepage for first 256 glyphs:
                    Do not reencode first 256 glyphs
        unchecked - Export only first 256 glyphs of the selected codepage
        unchecked - Put MS Char Set value into fsSelection field
        """
        return self._TTEExportUnicode

    @TTEExportUnicode.setter
    def TTEExportUnicode(self, value: int) -> None:
        self._TTEExportUnicode = value

    @property
    def TTEExportVOLT(self) -> int:
        return self._TTEExportVOLT

    @TTEExportVOLT.setter
    def TTEExportVOLT(self, value: int) -> None:
        self._TTEExportVOLT = value

    @property
    def TTEFontNames(self) -> int:
        """
        Do not export OpenType name records

        - 0 = Append OpenType names
        - 1 = Do not export OpenType names
        - 2 = Export only OpenType names
        """
        return self._TTEFontNames

    @TTEFontNames.setter
    def TTEFontNames(self, value: int) -> None:
        self._TTEFontNames = value

    @property
    def TTEHint(self) -> int:
        """
        Export hinted TrueType fonts?
        """
        return self._TTEHint

    @TTEHint.setter
    def TTEHint(self, value: int) -> None:
        self._TTEHint = value

    @property
    def TTEKeep(self) -> int:
        """
        Write stored TrueType native hinting
        """
        return self._TTEKeep

    @TTEKeep.setter
    def TTEKeep(self, value: int) -> None:
        self._TTEKeep = value

    @property
    def TTENoReorder(self) -> int:
        """
        Automatically reorder glyphs?
        """
        return self._TTENoReorder

    @TTENoReorder.setter
    def TTENoReorder(self, value: int) -> None:
        self._TTENoReorder = value

    @property
    def TTESubrize(self) -> int:
        return self._TTESubrize

    @TTESubrize.setter
    def TTESubrize(self, value: int) -> None:
        self._TTESubrize = value

    @property
    def TTEVisual(self) -> int:
        """
        Export visual TrueType hints?
        """
        return self._TTEVisual

    @TTEVisual.setter
    def TTEVisual(self, value: int) -> None:
        self._TTEVisual = value

    @property
    def TTEWriteBitmaps(self) -> int:
        """
        Export embedded bitmaps
        """
        return self._TTEWriteBitmaps

    @TTEWriteBitmaps.setter
    def TTEWriteBitmaps(self, value: int) -> None:
        self._TTEWriteBitmaps = value

    @property
    def TTEWriteKernFeature(self) -> int:
        """
        Generate OpenType "kern" feature if it is undefined or outdated?
        """
        return self._TTEWriteKernFeature

    @TTEWriteKernFeature.setter
    def TTEWriteKernFeature(self, value: int) -> None:
        self._TTEWriteKernFeature = value

    @property
    def TTEheadBBoxSavings(self) -> int:
        return self._TTEheadBBoxSavings

    @TTEheadBBoxSavings.setter
    def TTEheadBBoxSavings(self, value: int) -> None:
        self._TTEheadBBoxSavings = value

    @property
    def TTHHintingOptions(self) -> int:
        return self._TTHHintingOptions

    @TTHHintingOptions.setter
    def TTHHintingOptions(self, value: int) -> None:
        self._TTHHintingOptions = value

    @property
    def TTIAutohint(self) -> int:
        return self._TTIAutohint

    @TTIAutohint.setter
    def TTIAutohint(self, value: int) -> None:
        self._TTIAutohint = value

    @property
    def TTIConvert(self) -> int:
        return self._TTIConvert

    @TTIConvert.setter
    def TTIConvert(self, value: int) -> None:
        self._TTIConvert = value

    @property
    def TTIDecompose(self) -> int:
        return self._TTIDecompose

    @TTIDecompose.setter
    def TTIDecompose(self, value: int) -> None:
        self._TTIDecompose = value

    @property
    def TTIFontNames(self) -> int:
        return self._TTIFontNames

    @TTIFontNames.setter
    def TTIFontNames(self, value: int) -> None:
        self._TTIFontNames = value

    @property
    def TTIKeepHints(self) -> int:
        return self._TTIKeepHints

    @TTIKeepHints.setter
    def TTIKeepHints(self, value: int) -> None:
        self._TTIKeepHints = value

    @property
    def TTIReadBitmaps(self) -> int:
        return self._TTIReadBitmaps

    @TTIReadBitmaps.setter
    def TTIReadBitmaps(self, value: int) -> None:
        self._TTIReadBitmaps = value

    @property
    def TTIReadKernFeature(self) -> int:
        return self._TTIReadKernFeature

    @TTIReadKernFeature.setter
    def TTIReadKernFeature(self, value: int) -> None:
        self._TTIReadKernFeature = value

    @property
    def TTIReadOT(self) -> int:
        return self._TTIReadOT

    @TTIReadOT.setter
    def TTIReadOT(self, value: int) -> None:
        self._TTIReadOT = value

    @property
    def TTIScale1000(self) -> int:
        return self._TTIScale1000

    @TTIScale1000.setter
    def TTIScale1000(self, value: int) -> None:
        self._TTIScale1000 = value

    @property
    def TTIStoreTables(self) -> int:
        return self._TTIStoreTables

    @TTIStoreTables.setter
    def TTIStoreTables(self, value: int) -> None:
        self._TTIStoreTables = value

    @property
    def TTToolReverseShift(self) -> int:
        return self._TTToolReverseShift

    @TTToolReverseShift.setter
    def TTToolReverseShift(self, value: int) -> None:
        self._TTToolReverseShift = value

    @property
    def TracerCurveFit(self) -> float:
        return self._TracerCurveFit

    @TracerCurveFit.setter
    def TracerCurveFit(self, value: int) -> None:
        self._TracerCurveFit = value

    @property
    def TracerCurves(self) -> int:
        return self._TracerCurves

    @TracerCurves.setter
    def TracerCurves(self, value: int) -> None:
        self._TracerCurves = value

    @property
    def TracerExtremePoints(self) -> int:
        return self._TracerExtremePoints

    @TracerExtremePoints.setter
    def TracerExtremePoints(self, value: int) -> None:
        self._TracerExtremePoints = value

    @property
    def TracerPresetMode(self) -> int:
        return self._TracerPresetMode

    @TracerPresetMode.setter
    def TracerPresetMode(self, value: int) -> None:
        self._TracerPresetMode = value

    @property
    def TracerStraightenAngle(self) -> int:
        return self._TracerStraightenAngle

    @TracerStraightenAngle.setter
    def TracerStraightenAngle(self, value: int) -> None:
        self._TracerStraightenAngle = value

    @property
    def TracerTolerance(self) -> int:
        return self._TracerTolerance

    @TracerTolerance.setter
    def TracerTolerance(self, value: int) -> None:
        self._TracerTolerance = value

    @property
    def TrackingDistance(self) -> int:
        return self._TrackingDistance

    @TrackingDistance.setter
    def TrackingDistance(self, value: int) -> None:
        self._TrackingDistance = value

    @property
    def TrackingMode(self) -> int:
        return self._TrackingMode

    @TrackingMode.setter
    def TrackingMode(self, value: int) -> None:
        self._TrackingMode = value

    @property
    def UnicodeRangePercent(self) -> int:
        return self._UnicodeRangePercent

    @UnicodeRangePercent.setter
    def UnicodeRangePercent(self, value: int) -> None:
        self._UnicodeRangePercent = value

    @property
    def VendorCode(self) -> str:
        return self._VendorCode

    @VendorCode.setter
    def VendorCode(self, value: str) -> None:
        self._VendorCode = value

    # Not implemented in FontLab 4.5.2 Win:

    @property
    def CacheTTPath(self) -> str:  # unconfirmed str
        return self._CacheTTPath

    @CacheTTPath.setter
    def CacheTTPath(self, value: str) -> None:
        self._CacheTTPath = value

    @property
    def CacheTTUse(self) -> int:
        return self._CacheTTUse

    @CacheTTUse.setter
    def CacheTTUse(self, value: int) -> None:
        self._CacheTTUse = value

    @property
    def ContourSnapAllPoints(self) -> int:
        return self._ContourSnapAllPoints

    @ContourSnapAllPoints.setter
    def ContourSnapAllPoints(self, value: int) -> None:
        self._ContourSnapAllPoints = value

    @property
    def EditDeleteAlt(self) -> int:
        return self._EditDeleteAlt

    @EditDeleteAlt.setter
    def EditDeleteAlt(self, value: int) -> None:
        self._EditDeleteAlt = value

    @property
    def EditDoubleClickBitmap(self) -> int:
        return self._EditDoubleClickBitmap

    @EditDoubleClickBitmap.setter
    def EditDoubleClickBitmap(self, value: int) -> None:
        self._EditDoubleClickBitmap = value

    @property
    def EditEditSelection(self) -> int:
        return self._EditEditSelection

    @EditEditSelection.setter
    def EditEditSelection(self, value: int) -> None:
        self._EditEditSelection = value

    @property
    def EditNoToolbars(self):
        return self._EditNoToolbars

    @EditNoToolbars.setter
    def EditNoToolbars(self, value) -> None:
        self._EditNoToolbars = value

    @property
    def EditShowSelection(self) -> int:
        return self._EditShowSelection

    @EditShowSelection.setter
    def EditShowSelection(self, value: int) -> None:
        self._EditShowSelection = value

    @property
    def NamesFileName(self):
        return self._NamesFileName

    @NamesFileName.setter
    def NamesFileName(self, value) -> None:
        self._NamesFileName = value

    @property
    def OTReadMort(self) -> int:
        return self._OTReadMort

    @OTReadMort.setter
    def OTReadMort(self, value: int) -> None:
        self._OTReadMort = value

    @property
    def OTWriteMort(self) -> int:
        """
        Export "mort" table if possible?
        """
        return self._OTWriteMort

    @OTWriteMort.setter
    def OTWriteMort(self, value: int) -> None:
        self._OTWriteMort = value

    @property
    def ShowMeterPanel(self) -> int:
        return self._ShowMeterPanel

    @ShowMeterPanel.setter
    def ShowMeterPanel(self, value: int) -> None:
        self._ShowMeterPanel = value

    @property
    def UnicodeStrings(self) -> int:
        return self._UnicodeStrings

    @UnicodeStrings.setter
    def UnicodeStrings(self, value: int) -> None:
        self._UnicodeStrings = value

    # Methods

    def Init(self) -> None:
        """
        Reset FontLab Options to default settings
        """
        # These are not FL's default settings right now,
        # but sensible defaults set by an experienced user *cough*
        self.AutoAlignVector = 1
        self.AutohintingHRatio = 2.0
        self.AutohintingMaxHWidth = 0x00000082
        self.AutohintingMaxVWidth = 0x00000082
        self.AutohintingMinHLen = 0x0000001E
        self.AutohintingMinHWidth = 0x00000028
        self.AutohintingMinVLen = 0x0000001E
        self.AutohintingMinVWidth = 0x00000028
        self.AutohintingRemoveHints = 1
        self.AutohintingVRatio = 2.0
        self.AutoMetricsClose = 0x0000000A
        self.AutoMetricsLeft = 0x0000001E
        self.AutoMetricsRight = 0x0000001E
        self.AutoRemoveHints = 1
        self.AutoSave = 0
        self.AutoSaveTick = 0x0000000A
        self.Backup = 1
        self.BitmapSize = 0x000007D0
        self.CacheTTPath = ""
        self.CacheTTUse = 0
        self.ChartApplyTemplate = 0
        self.ChartAutoHide = 0
        self.ChartAutoUnicode = 1
        self.ChartCaptionFontCharSet = 0
        self.ChartCaptionFontName = "Geneva"
        self.ChartCaptionFontSize = 0x0000005A
        self.ChartDoubleClick = 1
        self.ChartDragDrop = 1
        self.ChartPreviewMarks = 1
        self.ChartSampleSize = 2
        self.ChartShowNotes = 1
        self.ChartSorting = 2
        self.ChartStyle = 0x000009F3
        self.ChartTemplateFontCharSet = 0
        self.ChartTemplateFontName = "Symbola"
        self.ChartTemplateFontSize = 0x000000F0
        self.ChartUseTemplate = 1
        self.ClosepathArrowLen = 0x00000008
        self.CodepagesCount = 0x00000098
        self.ColorBitmap = 0xFFB8B8C8
        self.ColorClosepath = 0x00C0C0C0
        self.ColorEcho = 0xFF8C8C8C
        self.ColorHintsPen = 0xFF008000
        self.ColorMaskPen = 0x01CE753E
        self.ColorSeacPen = 0xFF808080
        self.ColorTemplate = 0xFFFFC0C0
        self.ColorVMetrics = 0xFF808080
        self.ContourSnapAllPoints = 0
        self.CopyHDMXData = 1  # Copy HDMX data from base to composite glyph
        self.CreateUnexistingCharacters = 0
        self.DefaultGlyph = 0x00000041
        self.DuplicateX = 0x00000064
        self.DuplicateY = 0x00000064
        self.EditBCPsFixed = 1
        self.EditBitmapStyle = 0
        self.EditChangeCursor = 1
        self.EditDeleteAlt = 0
        self.EditDoubleClickBitmap = 1
        self.EditEditSelection = 0
        self.EditFollowScroll = 1
        self.EditGridX = 0x00000032
        self.EditGridY = 0x00000032
        self.EditHandleTool = 1
        self.EditHitDistance = 3
        self.EditInstantRefresh = 0
        self.EditLeaveEcho = 1
        self.EditNotFillOpen = 1
        # self.EditNoToolbars = value
        self.EditRulers = 1
        self.EditScaleEPS = 0
        self.EditShowCross = 1
        self.EditShowPosition = 3
        self.EditShowSelection = 1
        self.EditSmallNodes = 0
        self.EditSmoothOutline = 1
        self.EraseSize = 5
        self.FitAscender = 0x000003E8
        self.FitDescender = 0xFFFFFE70
        self.FontAudit = 0xFFFF7FFF
        self.GlyphsBarOpen = 1
        self.HideAllLayers = 0
        self.HideToolbars = 0
        self.LockStyleEx = 0
        # self.NamesFileName = value
        self.OptimizeAlign = 1
        self.OptimizeReduce = 1
        self.OTAddClasses = 1
        self.OTReadMort = 0
        self.OTWriteGDEF = 0
        self.OTWriteMort = 0
        self.OverlapMode = 2
        self.PaintAutoTransform = 0
        self.PaintAutoView = 0
        self.PaintBrushBody = 0
        self.PaintBrushCap = 0
        self.PaintBrushJoin = 1
        self.PaintBrushRoundness = 0x0320
        self.PaintBrushVectorX = 0x00000068
        self.PaintBrushVectorY = 0x000003E2
        self.PaintBrushWidth = 0x00000078
        self.PaintColor = 3
        self.PaintMode = 2
        self.PaintStyleEx = 0x30849893
        self.PaintTextSize = 0x0000012C
        self.PaintTool = 6
        self.PaintTracePolygon = 0
        self.PasteX = 0
        self.PasteY = 0
        self.PreviewExpanded = 1
        self.PreviewExpandedHeight = 0x000003E8
        self.PreviewPointSize = "5-36, 38, 40, 42, 48"
        self.PreviewPPMs = "5-36, 38, 40, 42, 48"
        self.PreviewPPMsExpanded = "8-40"
        self.PreviewSecondLine = "Tonn Hauvonid DHLYENO"
        self.PreviewSmooth = 0
        self.PreviewWidth = 0x0000033B
        self.SamplePPM1 = 0x07FFFFC0
        self.SamplePPM2 = 0
        self.ShowMeterPanel = 1
        self.SnapStyleEx = 0xF
        self.T1AFM = 0
        self.T1Autohint = 0
        self.T1Decompose = 0
        self.T1Encoding = 1
        self.T1PFM = 0
        self.T1Terminal = 0
        self.T1Unicode = 1
        self.T1UseOS2 = 0
        self.TracerCurveFit = 0x00004E20
        self.TracerCurves = 1
        self.TracerExtremePoints = 1
        self.TracerPresetMode = 0
        self.TracerStraightenAngle = 3
        self.TracerTolerance = 3
        self.TrackingDistance = 0x32
        self.TrackingMode = 0
        self.TTEAddCharacters = 0
        self.TTEApplyBBoxSavings = 1
        self.TTEAutohint = 0
        self.TTEAutoWinAscDesc = 1
        self.TTECmap10 = 0
        self.TTEExportOT = 1
        self.TTEExportUnicode = 1
        self.TTEExportVOLT = 0
        self.TTEFontNames = 2
        self.TTEheadBBoxSavings = 0x40
        self.TTEHint = 1
        self.TTEKeep = 0
        self.TTENoReorder = 1
        self.TTESubrize = 1
        self.TTEVisual = 1
        self.TTEWriteBitmaps = 0
        self.TTEWriteKernFeature = 0
        self.TTHHintingOptions = 0x00001107
        self.TTIAutohint = 0
        self.TTIConvert = 0
        self.TTIDecompose = 0
        self.TTIFontNames = 2
        self.TTIKeepHints = 1
        self.TTIReadBitmaps = 1
        self.TTIReadKernFeature = 1
        self.TTIReadOT = 1
        self.TTIScale1000 = 0
        self.TTIStoreTables = 0
        self.TTToolReverseShift = 1
        self.UnicodeRangePercent = 0
        self.UnicodeStrings = 1
        self.VendorCode = "pyrs"

        # Not documented

        self.ATMPPM = 0x0000002A
        self.ATMScaler = 0x61747375
        self.ATMSmooth = 1
        self.AutohintingEngine = 0
        self.ChartAutoActivateCodepage = 1
        self.ChartCopyKerning = 0
        self.ChartGenerate = 1
        self.ChartGenerateRTL = 0
        self.ColorBackground = 0xFFFFFFFF
        self.ColorBackgroundMask = 0xFFFFF0FF
        self.ColorBackgroundMetrics = 0xFFFFFFF0
        self.ColorForegroundMetrics = 0xFF000000
        self.ColorGlobalGuide = 0xFFFF0000
        self.ColorGrid = 0xFFE7E7E7
        self.ColorGroups = 0xFFE0E0E0
        self.ColorGuide = 0xFF0000FF
        self.ColorNeighbors = 0xFF404040
        self.ColorOutline = 0xFF000000
        self.ColorTemplate = 0xFFFFC0C0
        self.ControlSwitch = 0
        self.CustomDict = "*"
        self.DSIG_KeyPath = ""
        self.DSIG_Password = ""
        self.DSIG_SertPath = ""
        self.DSIG_TimeStamp = 0
        self.DSIG_Use = 0  # Generate digital signature (DSIG table)
        self.EditAllSmooth = 1
        self.EditAutoSelectLayers = 1
        self.EditBCVOpacity = 0x00000032
        self.EditComponentsByOutline = 0
        self.EditConnectMasters = 0
        self.EditDoubleClickMask = 1
        self.EditInstantRefresh = 0
        self.EditKeyboardEditBCP = 1
        self.EditLayersExpanded = 1
        self.EditPreviewKey = 0x000000C0
        self.EditShowAnchorNames = 1
        self.EditShowMasterPoints = 1
        self.EditShowMeasurementLine = 1
        self.EditShowNodesSelection = 1
        self.EditSmoothIsSmooth = 1
        self.EditThickOutline = 0
        self.EditUndoSelection = 0
        self.EditZoomMode = 0x00000007
        self.EmbeddingType = 0x00000004
        self.EnableExtPythonEditor = 1
        self.EnablePython = 1
        self.EnableStickPanels = 0
        self.EnableTooltips = 1
        self.ExpandKernCodepage = "MS Windows 1252 Western (ANSI)"
        self.ExpandKernCount = 0x00002AA7
        self.ExpandKernFlags = 0x00000012
        self.ExtPythonEditor = "/Applications/TextMate.app"
        self.FileOpenSample = "ABRaeg123"
        self.FontAuditLive = 0
        self.FontDialogName = "Lucida Grande"
        self.FontDialogSize = 0x0000005A
        self.FontExpanded = 1
        self.FontFixedName = "Lucida Grande"
        self.FontFixedSize = 0x00000050
        self.FontLabNodes = 0
        self.FontPanelName = "Lucida Grande"
        self.FontPanelSize = 0x0000005A
        self.FontRulerName = "Lucida Grande"
        self.FontRulerSize = 0x0000005A
        self.FontSmallName = ".TiniNumbers"
        self.FontSmallSize = 0x00000046
        self.FontTableName = "Lucida Grande"
        self.FontTableSize = 0x00000050
        self.GlyphNameSortMode = 1
        self.GlyphsBarTop = 1
        self.GroupCenter = 0
        self.GroupDoubleClick = 1
        self.GroupFill = 0
        self.GroupKerning = 1
        self.GroupMask = 0
        self.GroupShapeFill = 0
        self.GroupShapeMask = 0
        self.GroupShapeOpacity = 0x00000019
        self.GroupShiftX = 0
        self.GroupShiftY = 0
        self.GroupUseMetrics = 0
        self.MacroFontName = "Andale Mono"
        self.MaskMetrics = 1
        self.MetricsBarExpanded0 = 0
        self.MetricsBarExpanded1 = 1
        self.MetricsBarExpanded2 = 1
        self.MetricsBarExpanded3 = 1
        self.MetricsBottomBar0 = 0
        self.MetricsBottomBar1 = 1
        self.MetricsBottomBar2 = 1
        self.MetricsBottomBar3 = 1
        self.MetricsShowBar0 = 1
        self.MetricsShowBar1 = 1
        self.MetricsShowBar2 = 1
        self.MetricsShowBar3 = 1
        self.MetricsShowTable0 = 0
        self.MetricsShowTable1 = 1
        self.MetricsShowTable2 = 1
        self.MetricsShowTable3 = 1
        self.MMExtrapolation = 0
        self.OpenOutput = 1
        self.OTAddMetricsClasses = 0
        self.OTCompileFeatures = 1
        self.OTGenerate = 0
        self.OTOldContextRule = 0
        self.OTPreviewSize = 0x00000080
        self.OTSampleSize = 0x00000060
        self.OutputFontName = "Andale Mono"
        self.PaintJapanese = 0
        self.PaintReverseBrushAutoColor = 0
        self.PaintTextFont = "Arial"
        self.PreviewAlternativeArrows = 0
        self.PreviewApplyTemplate = 1
        self.PreviewAutoScroll = 1
        self.PreviewAutoWrap = 1
        self.PreviewBlueColor = 0xFF1A1D1E
        self.PreviewFilterKerning = 0
        self.PreviewFilterMetrics = 1
        self.PreviewFocusString = 1
        self.PreviewHighlightKeyGlyphs = 0
        self.PreviewKeepExceptions = 1
        self.PreviewKerningIcons = 1
        self.PreviewListWidth = 0x000000C8
        self.PreviewLockDependences = 1
        self.PreviewLockStyle = 0
        self.PreviewMode = 2
        self.PreviewPaintStyle = 0x00000100
        self.PreviewPanelTop = 1
        self.PreviewProcessFeatures = 0
        self.PreviewShowClassMembers = 1
        self.PreviewShowReferences = 1
        self.PreviewSize = 0x000000A0
        self.PreviewSnapStyle = 0x00000010
        self.QuickZoomScale = 0x00001D4C
        self.RemoveZeroKerning = 0
        self.Shift10 = 0x0000000A
        self.StyleInactiveMetrics = 2
        self.StyleMetrics = 1
        self.SyncronizeMasters = 1
        self.T1ExportCompatibleCyrillic = 0
        self.T1ExportEncoding = 0
        self.T1FSType = 0
        self.T1MatchEncoding = 0
        self.T1UseOTFamilyName = 0
        self.T1UseOTStyleName = 0
        self.T1UseTrademarkName = 0
        self.Template_BDF = ""
        self.Template_Path = ""
        self.Template_Use = 0
        self.TTEDecompose = 1
        self.TTESmartFontName = 0
        self.TTESmartMacNames = 1  # Use the OpenType names as menu names on Macintosh
        self.TTEStoreTables = 0  # Export OpenType layout tables
        self.TTEVersionOS2 = 3  # OS/2 table version 3
        self.TTEVisual = 1
        self.TTEWriteKernTable = 0  # Export old-style non-OpenType "kern" table
        self.TTFScaler = 0x6D737363
        self.TTFSmooth = 0
        self.TTHPPM = 0x00000013
        self.TTIGenerateNames = 1
        self.TTIReadBinaryOT = 1
        self.TTIStoreBinaryOT = 1
        self.UnicodeKeyboard = 0
        self.UnicodeRangePercent = 0

    def Load(self) -> None:
        """
        Read FontLab Options from registry
        """
        # Synonymous to Init()? As long as Init() doesn't actually set the
        # default values ...
        self.Init()

    def Save(self) -> None:
        """
        Save FontLab Options in registry
        """
        pass



from __future__ import annotations

from typing import TYPE_CHECKING

from FL.fake.Base import Copyable
from FL.objects.Point import Point

if TYPE_CHECKING:
    from FL.objects import Glyph
    from FL.objects.Matrix import Matrix


class Anchor(Copyable):
    """
    Anchor - class to represent Anchor point
    """

    __slots__ = ["_mark", "_name", "_parent", "_points", "_reserved"]

    # Constructor

    def __init__(
        self, anchor_or_name: Anchor | str | None = None, x: int = 0, y: int = 0
    ) -> None:
        """
        Integer values are accepted as coordinates.

        Args:
            anchor_or_name (Anchor | str | None, optional): An `Anchor` to copy data
                from, or the anchor's name as a string. Defaults to None.
            x (int, optional): Horizontal position of the anchor. Defaults to 0.
            y (int, optional): Vertical position of the anchor. Defaults to 0.
        """
        self._set_defaults()

        # Process params

        if isinstance(anchor_or_name, Anchor):
            self._copy_constructor(anchor_or_name)

        elif isinstance(anchor_or_name, str):
            self.name = anchor_or_name
            self.x = x
            self.y = y

    def __repr__(self) -> str:
        if self._parent is None:
            return f"<Anchor: name={self.name}, x={self.x}, y={self.y}, orphan>"
        return f"<Anchor: name={self.name}, x={self.x}, y={self.y}, active refernce>"

    def _set_defaults(self) -> None:
        self._parent = None
        self._name = ""

        self._mark = 1
        self._reserved = 0
        self._points = [Point() for _ in range(16)]

    # Attributes

    @property
    def parent(self) -> Glyph | None:
        """
        Anchors's parent object, `Glyph`
        """
        return self._parent

    @property
    def name(self) -> str:
        """
        Anchor's name as a string

        Returns:
            str: The name
        """
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @property
    def x(self) -> int:
        """
        Horizontal position of the anchor

        Returns:
            int: The x coordinate
        """
        return int(self._points[0].x)

    @x.setter
    def x(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError
        for p in self._points:
            p.x = value

    @property
    def y(self) -> int:
        """
        Vertical position of the anchor

        Returns:
            int: The y coordinate
        """
        return int(self._points[0].y)

    @y.setter
    def y(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError
        for p in self._points:
            p.y = value

    @property
    def p(self) -> Point:
        """
        Position of the anchor as a `Point` object

        Returns:
            Point: The point
        """
        return self._points[0]

    @p.setter
    def p(self, value: Point) -> None:
        # Truncates any floats
        x = int(value.x)
        y = int(value.y)
        for p in self._points:
            p.x = x
            p.y = y

    @property
    def mark(self) -> int:
        """
        Behaves like the `mark` attribute of the `Glyph` object

        Returns:
            int: The hue in degrees of the mark color
        """
        return self._mark

    @mark.setter
    def mark(self, value: int) -> None:
        # TODO: clamp to 0 to 65535
        # actually: -1 -> 65535, 65536 -> 0 etc.
        self._mark = value

    # Operations

    # Anchor has no operations

    # Methods

    def Layer(self, masterindex: int) -> Point:
        """
        Returns point for the master 'masterindex'.
        """
        return self._points[masterindex]

    def SetLayer(self, masterindex: int, point: Point) -> None:
        """
        Set the point for the master 'masterindex'.
        """
        pt = self._points[masterindex]
        pt.x = int(point.x)
        pt.y = int(point.y)

    def Transform(self, m: Matrix) -> None:
        """
        Applies Matrix transformation to the Anchor (see `Matrix`)

        Args:
            m (Matrix): The transformation matrix
        """
        p = self.p
        m.fake_transform_point(p)
        self.x = int(p.x)  # TODO: Is truncation to int correct?
        self.y = int(p.y)  # TODO: Is truncation to int correct?



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any

from FL.fake.Font import FakeFont
from FL.helpers.classList import ClassList
from FL.helpers.ListParent import ListParent
from FL.objects.Encoding import Encoding
from FL.objects.Rect import Rect
from FL.objects.TTInfo import TTInfo
from FL.objects.Uni import Uni
from FL.objects.WeightVector import WeightVector

if TYPE_CHECKING:
    from vfbLib.typing import CustomCmap

    from FL.objects.EncodingRecord import EncodingRecord
    from FL.objects.Feature import Feature
    from FL.objects.Glyph import Glyph
    from FL.objects.Guide import Guide
    from FL.objects.NameRecord import NameRecord
    from FL.objects.TrueTypeTable import TrueTypeTable


class Font(FakeFont):
    """
    Base class to represent a font
    """

    __slots__ = [
        "_axis",
        "_classes",
        "_encoding",
        "_features",
        "_file_name",
        "_fontnames",
        "_glyphs",
        "_license_url",
        "_license",
        "_masters_count",
        "_postscript_hinting_options",
        "_truetypetables",
        "_unknown_pleasures",
        "_xuid_num",
        "_xuid",
        "apple_name",
        "ascender",
        "blue_fuzz",
        "blue_scale",
        "blue_shift",
        "blue_values_num",
        "blue_values",
        "cap_height",
        "codepages",
        "copyright",
        "customdata",
        "default_character",
        "default_width",
        "descender",
        "designer_url",
        "designer",
        "family_blues_num",
        "family_blues",
        "family_name",
        "family_other_blues_num",
        "family_other_blues",
        "fond_id",
        "font_name",
        "font_style",
        "force_bold",
        "full_name",
        "hguides",
        "is_fixed_pitch",
        "italic_angle",
        "mac_compatible",
        "menu_name",
        "modified",
        "ms_charset",
        "ms_id",
        "note",
        "notice",
        "ot_classes",
        "other_blues_num",
        "other_blues",
        "panose",
        "pcl_chars_set",
        "pcl_id",
        "pref_family_name",
        "pref_style_name",
        "slant_angle",
        "source",
        "stem_snap_h_num",
        "stem_snap_h",
        "stem_snap_v_num",
        "stem_snap_v",
        "style_name",
        "trademark",
        "tt_u_id",
        "tt_version",
        "ttinfo",
        "underline_position",
        "underline_thickness",
        "unicoderanges",
        "unique_id",
        "upm",
        "vendor_url",
        "_vendor",
        "version_major",
        "version_minor",
        "version",
        "vguides",
        "vp_id",
        "weight_code",
        "weight_vector",
        "weight",
        "width",
        "x_height",
        "year",
        # Non-API:
        "_custom_cmaps",
        "_encoding_default",
        "_export_pclt_table",
        "_export_options",
        "_ot_export_options",
        "_pclt_table",
        "_axis_count",
        "_anisotropic_interpolation_mappings",
        "_axis_mappings_count",
        "_axis_mappings",
        "_kerning_class_flags",
        "_master_names",
        "_master_locations",
        "_master_ps_infos",
        "_mapping_mode",
        "_metrics_class_flags",
        "_primary_instance_locations",
        "_primary_instances",
        # Internal:
        "fake_vfb_object",
    ]

    # Constructor

    def __init__(
        self,
        font_or_path: Font | str | None = None,
        instances: tuple[float, float, float, float] | None = None,
    ) -> None:
        super().__init__()
        self._set_defaults()

        # Process params

        if isinstance(font_or_path, Font):
            if instances is None:
                self._copy_constructor(font_or_path)
            else:
                # Generate an instance
                # instances is a tuple containing instance values for all MM
                # axes defined in the font
                raise NotImplementedError
        elif isinstance(font_or_path, str) or isinstance(font_or_path, Path):
            # Instantiate with path
            self.Open(font_or_path)

        # else: Empty font

    def __repr__(self) -> str:
        return "<Font: '%s', %i glyphs>" % (self.full_name, len(self))

    # Attributes

    @property
    def file_name(self) -> str:
        """
        Full path of the file from which the font was opened/saved.
        """
        # The file path is stored internally as Path, but we return a str
        return str(self._file_name)

    @property
    def axis(self) -> list[tuple[str, str, str]]:
        """
        Array of font axes. Not reported by docstring nor e-font.

        Example: [('Weight', 'Wt', 'Weight')]
        """
        return self._axis

    @property
    def classes(self) -> list[str]:
        """
        List of glyph classes.

        Returns a copy of the list. To change entries, you must reassign the whole list.
        """
        return list(self._classes)

    @classes.setter
    def classes(self, value: list[str]) -> None:
        self._classes.fake_set_classes(value)

    @property
    def encoding(self) -> Encoding:
        """
        Current encoding of the font.
        """
        return self._encoding

    @property
    def features(self) -> ListParent[Feature]:
        """
        List of Opentype features.
        """
        return self._features

    @features.setter
    def features(self, value: list[Feature]) -> None:
        raise RuntimeError("Class Font has no attribute features or it is read-only")

    @property
    def fontnames(self) -> ListParent[NameRecord]:
        """
        List of font name records.
        """
        return self._fontnames

    @fontnames.setter
    def fontnames(self, value: list[NameRecord]) -> None:
        raise RuntimeError

    @property
    def glyphs(self) -> ListParent[Glyph]:
        """
        Return the array of glyphs.
        """
        # Read-only.
        return self._glyphs

    @glyphs.setter
    def glyphs(self, value: list[Glyph]) -> None:
        raise RuntimeError

    @property
    def truetypetables(self) -> ListParent[TrueTypeTable]:
        """
        List of custom TrueType tables.
        """
        # Read-only.
        return self._truetypetables

    @truetypetables.setter
    def truetypetables(self, value: list[TrueTypeTable]) -> None:
        raise RuntimeError

    @property
    def vendor(self) -> str:
        return self._vendor

    @vendor.setter
    def vendor(self, value: str) -> None:
        if len(value) > 4:
            self._vendor = value[:4]
        else:
            self._vendor = value

    @property
    def xuid(self) -> list[int]:
        """
        A list of Type 1 XUID numbers.

        The list may contain from 0 to 20 entries. To change the length of the list, set
        `Font.xuid_num`.


        Returns:
            list[int]: The list of Type 1 XUID numbers.
        """
        return self._xuid

    @property
    def xuid_num(self) -> int:
        """_summary_

        Returns:
            int: _description_
        """
        return self._xuid_num

    @xuid_num.setter
    def xuid_num(self, value: int) -> None:
        if not 0 <= value <= 20:
            raise RuntimeError('New "xuid_num" is out of range 0..20')
        diff = value - self._xuid_num
        if diff < 0:
            self._xuid = self.xuid[:value]
        elif diff > 0:
            self._xuid.extend([0] * diff)
        self._xuid_num: int = value

    # Operations

    def __len__(self) -> int:
        """
        Return the number of glyphs.

        Returns:
            int: The number of glyphs.
        """
        return len(self._glyphs)

    def __getitem__(self, index: int | str) -> Glyph | None:
        """
        Access the glyphs array.

        Args:
            index (int | str): The glyph index to get.

        Returns:
            Glyph | None: Return the glyph, or None if the index was out of bounds.
        """
        if isinstance(index, str):
            # We got a glyph name
            for glyph in self._glyphs:
                if glyph.name == index:
                    return glyph
            return None

        return self._glyphs[index]

    # Methods

    def New(self) -> None:
        """
        Clear the font.
        """
        raise NotImplementedError

    def Open(self, filename: str) -> int:
        """
        Open a font from a VFB file.

        Args:
            filename (str): The path and file name of the VFB file.

        Returns:
            int: 1 on success, 0 if the file could not be opened.

        If you need to import a font (not in VFB format), use `FL.Open()` or
        `FL.OpenFont()`.
        """
        from FL.vfb.reader import VfbToFontReader

        self._set_file_name(None)  # TODO: What if the font already is loaded from disk?
        try:
            reader = VfbToFontReader(Path(filename), self)
            self.fake_vfb_object = reader.vfb
        except:  # noqa: E722
            raise
            return 0
        self._set_file_name(filename)
        return 1

    def Save(self, filename: str) -> None:
        """
        Save the font in VFB format.

        Args:
            filename (str): The path and file name of the VFB file.
        """
        from FL.vfb.writer import FontToVfbWriter

        self._set_file_name(filename)
        FontToVfbWriter(self, Path(filename))

    def OpenAFM(self, filename: str, mode: int, layer: int) -> None:
        """
        open AFM-File, mode is the integer bit field.
          The bit list is:
          ALLMETRICS       - 0x0001
          THICKERMETRICS   - 0x0002
          WIDERMETRICS     - 0x0004
          CLOSEMETRICS     - 0x0008
          REPLACEKERNING   - 0x0010
          ADDKERNING       - 0x0020
          AUTOKERNING      - 0x0040
          REPLACEOTHERDATA - 0x0100
          REPLACENAMES     - 0x0200

          Constants for mode (only in FL 4.5 Mac)
          mtALLMETRICS
          mtTHICKERMETRICS
          mtWIDERMETRIC
          mtCLOSEMETRICS
          mtREPLACEKERNING
          mtADDKERNING
          mtAUTOKERNING
          mtREPLACEOTHERDATA
          mtREPLACENAMES
        """
        raise NotImplementedError

    def SaveAFM(self, filename: str) -> None:
        """
        Save an AFM and an INF file.

        Args:
            filename (str): The path and filename to save the files to.
        """
        afm = self.fake_get_afm()
        afm_path = Path(filename).with_suffix(".afm")
        with open(afm_path, "w") as f:
            f.write(afm)
        inf = self.fake_get_inf()
        inf_path = Path(filename).with_suffix(".inf")
        with open(inf_path, "w") as f:
            f.write(inf)

    def _normalize_upm(self, value: float) -> int:
        # TODO: truncate value before scaling?
        return int(value * 1000 / self.upm)

    def fake_bounding_rect(self, for_afm: bool = False) -> Rect:
        if for_afm:
            x0 = 0
            y0 = self.descender[0] - (100 * self.upm / 1000)
            y1 = self.ascender[0] + (100 * self.upm / 1000)
        else:
            x0 = 32767
            y0 = 32767
            y1 = -32767
        rect = Rect(x0, y0, -32767, y1)
        for g in self.glyphs:
            gr = g.GetBoundingRect()
            # print(g.name, gr)
            rect += gr
        return rect

    def fake_save_afm_expanded(self, filename: str) -> None:
        afm = self.fake_get_afm(expand_kerning=True)
        afm_path = Path(filename).with_suffix(".afm")
        with open(afm_path, "w") as f:
            f.write(afm)

    def fake_get_afm(self, expand_kerning: bool = False) -> str:
        afm = ["StartFontMetrics 2.0"]
        r = self.fake_bounding_rect(for_afm=True)
        bbox = (
            f"{self._normalize_upm(r.ll.x)} {self._normalize_upm(r.ll.y)} "
            f"{self._normalize_upm(r.ur.x)} {self._normalize_upm(r.ur.y)}"
        )
        if bbox == "32767 32767 -32767 -32767":
            bbox = "0 0 0 0"
        afm.extend(
            [
                f"Comment Copyright {self.notice}",
                f"Comment Panose {' '.join([str(p) for p in self.panose])}",
                f"FullName {self.full_name}",
                f"FontName {self.font_name}",
                f"FamilyName {self.family_name}",
                f"Weight {self.weight}",
                f"Notice {self.copyright}",
                f"Version {self.version_major}.{self.version_minor:03d}",
                f"IsFixedPitch {('false', 'true')[self.is_fixed_pitch]}",
                f"ItalicAngle {self.italic_angle:0.2f}",
                f"FontBBox {bbox}",
                f"Ascender {self._normalize_upm(self.ascender[0])}",
                f"Descender {self._normalize_upm(self.descender[0])}",
                f"XHeight {self._normalize_upm(self.x_height[0])}",
                f"CapHeight {self._normalize_upm(self.cap_height[0])}",
                f"UnderlinePosition {self._normalize_upm(self.underline_position)}",
                f"UnderlineThickness {self._normalize_upm(self.underline_thickness)}",
                "EncodingScheme FontSpecific",
                f"StartCharMetrics {len(self.glyphs)}",
            ]
        )
        glyphs = self.fake_sort_glyphs(self.glyphs.data)
        for g in glyphs:
            r = g.bounding_box
            bbox = (
                f"{self._normalize_upm(r.ll.x)} {self._normalize_upm(r.ll.y)} "
                f"{self._normalize_upm(r.ur.x)} {self._normalize_upm(r.ur.y)}"
            )
            if bbox == "32767 32767 -32767 -32767":
                bbox = "0 0 0 0"
            afm.append(
                f"C {g.unicode or -1} ; WX {self._normalize_upm(g.width)} ; "
                f"N {g.name} ; B {bbox} ;"
            )
        afm.append("EndCharMetrics")

        kerning = self.fake_get_afm_kerning(expand_kerning)
        if kerning:
            kerning = self.fake_sort_kerning(kerning)
            afm.append("StartKernData")
            afm.append(f"StartKernPairs {len(kerning)}")
            prev_L = ""
            for L, R, value in kerning:
                if prev_L != "" and prev_L != L:
                    afm.append("")
                afm.append(f"KPX {L} {R} {self._normalize_upm(value)}")
                prev_L = L

            afm.append("")
            afm.append("EndKernPairs")
            afm.append("EndKernData")
        afm.append("EndFontMetrics\n")

        return "\n".join(afm)

    def fake_get_afm_kerning(
        self, expand_kerning: bool = False
    ) -> list[tuple[str, str, int]]:
        if expand_kerning:
            self.fake_kerning.expand()
            return self.fake_kerning.flat_kerning

        kerning = []
        for g in self.glyphs:
            L = g.name
            for pair in g.kerning:
                R = self.glyphs[pair.key].name
                value = pair.value
                kerning.append((L, R, value))
        return kerning

    def fake_get_inf(self) -> str:
        inf = ""
        return inf

    def fake_sort_glyphs(self, glyphs: list[Glyph]) -> list[Glyph]:
        glyph_order = tuple([rec.name for rec in self.encoding])
        sortable = [(glyph_order.index(glyph.name), glyph) for glyph in glyphs]
        sortable.sort()
        # print(sortable)
        return [glyph for _, glyph in sortable]

    def fake_sort_kerning(
        self, kerning: list[tuple[str, str, int]]
    ) -> list[tuple[str, str, int]]:
        glyph_order = tuple([rec.name for rec in self.glyphs])
        sortable = [
            (glyph_order.index(L), glyph_order.index(R), L, R, value)
            for L, R, value in kerning
        ]
        sortable.sort()
        return [(L, R, value) for _, _, L, R, value in sortable]

    def Reencode(self, e: Encoding, style: int = 0) -> None:
        """
        Apply an encoding to the font.

        The parameters of this method are not reported by the docstring and I don't know
        what the style parameter does.

        Args:
            e (Encoding): The encoding.
            style (int, optional): _description_. Defaults to 0.
        """
        raise NotImplementedError

    def FindGlyph(self, name_uni_int: str | Uni | int) -> int:
        """
        (name: str) | (unicode: Uni) | (unicode: int)
        - finds glyph and return its index or -1
        """
        if isinstance(name_uni_int, str):
            # name
            for i, g in enumerate(self._glyphs):
                if g.name == name_uni_int:
                    return i
            return -1
        elif isinstance(name_uni_int, Uni):
            # uni object
            for i, g in enumerate(self._glyphs):
                if name_uni_int.value in g.unicodes:
                    return i
            return -1
        elif isinstance(name_uni_int, int):
            # int (unicode value)
            for i, g in enumerate(self._glyphs):
                if name_uni_int in g.unicodes:
                    return i
            return -1
        else:
            raise TypeError

    def DefineAxis(self, name: str, type: str, shortname: str) -> None:
        """
        Defines a new Multiple Master axis.

        Args:
            name (str): _description_
            type (str): _description_
            shortname (str): _description_
        """
        raise NotImplementedError

    def DeleteAxis(self, axisindex: int, position: float) -> None:
        """
        Removes the axis

        Args:
            axisindex (int): _description_
            position (float): _description_
        """
        raise NotImplementedError

    def GenerateUnicode(self) -> None:
        """
        Generates Unicode indexes for all glyphs
        """
        raise NotImplementedError

    def GenerateNames(self) -> None:
        """
        Generates names for all glyphs
        """
        raise NotImplementedError

    def GenerateGlyph(self, name: str) -> Glyph:
        """
        Generates new glyph using 'name' as a source of information about glyph's
        composition. See 'FontLabDir/Mapping/alias.dat' for composition definitions.

        Args:
            name (str): _description_

        Returns:
            Glyph: _description_

        The glyph is not added to the font automatically.
        """
        glyph = Glyph()
        glyph.name = name
        return glyph

    def has_key(self, name_uni_int: str | Uni | int) -> int:
        """
        Find a glyph by name, unicode or integer unicode and return 1 (found) or 0 (not
        found).

        Args:
            name_uni_int (str | Uni | int): _description_

        Returns:
            int: 1 if the glyph name or unicode value are present in the font,
                0 otherwise.
        """
        glyph_index = self.FindGlyph(name_uni_int)
        if glyph_index == -1:
            return 0
        return 1

    def GenerateFont(self, fontType: int, filename: str) -> None:
        """
        Generate a font. Deprecated. See the `FL.objects.FontLab` class for a
        description.

        As a method of the `Font` class, this method is deprecated. Since FontLab 4.52
        for Mac and FontLab 4.53 for Windows, `GenerateFont` is a method of the
        `FontLab` class.

        Args:
            fontType (int): The font type.
            filename (str): The path and file name of the generated font.

        Raises:
            AttributeError: In FontLab 5, the method is deprecated.
        """
        raise AttributeError

    # Undocumented methods

    def MakeKernFeature(self, vector: WeightVector) -> None:
        """
        Generates 'kern' feature using font kerning and classes

        Args:
            vector (WeightVector): The `WeightVector` used to interpolate the kerning
                values.
        """
        raise NotImplementedError

    def MergeFonts(self, source: Font, flags: int | None = None) -> None:
        """
        Append all glyphs from the source font to the current fonts.
        Check mfXXXX constants for options.

        Args:
            source (Font): The source font to be merged.
            flags (int | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def SetClassFlags(
        self,
        class_index: int,
        left_lsb: bool | int,
        right_rsb: bool | int,
        width: bool | int | None = None,
    ) -> None:
        """
        (int class_index, bool left, bool right)
        - allows to set 'left' and 'right' properties of the kerning class

        (int class_index, bool lsb, bool rsb, bool width)
        - allows to set 'lsb', 'rsb' and 'width' properties of the metrics class
        """
        self._classes.SetClassFlags(class_index, left_lsb, right_rsb, width)

    def GetClassLeft(self, class_index: int) -> int | None:
        """
        Return the 'left' property of the class.

        Args:
            class_index (int): _description_

        Returns:
            int | None: Returns 0 for non-kerning classes, 0 or 1 for kerning classes,
                None for class_index outside the classes list length.
        """
        return self._classes.GetClassLeft(class_index)

    def GetClassRight(self, class_index: int) -> int | None:
        """
        Return the 'right' property of the class.

        Args:
            class_index (int): _description_

        Returns:
            int | None: Returns 0 for non-kerning classes, 0 or 1 for kerning classes,
                None for class_index outside the classes list length.
        """
        return self._classes.GetClassRight(class_index)

    def GetClassMetricsFlags(self, class_index: int) -> tuple | None:
        """
        (int class_index)
        - returns the tuple containing LSB, RSB and Width flags of the metrics
        class
        """
        return self._classes.GetClassMetricsFlags(class_index)

    # Additional methods reported by dir(fl.font)

    def GenerateInstance(self) -> None:
        raise NotImplementedError

    # Defaults

    def fake_clear_defaults(self) -> None:
        """
        Clear some lists prior to deserializing a font from a Vfb.
        """
        self._master_names.clear()
        self._master_locations.clear()
        self._master_ps_infos.clear()

    def _set_defaults(self) -> None:
        # Identification

        # full path of the file from which the font was opened/saved
        self._file_name = None
        # font Family Name
        self.family_name: str | None = None
        # font Style Name
        self.style_name: str | None = None
        # font Full Name
        self.full_name: str | None = None
        # font Font Name
        self.font_name: str | None = None

        # Font Style as BitList:
        # italic       =  1
        # underscored  =  2
        # negative     =  4
        # outlined     =  8
        # strikethough = 16
        # bold         = 32
        self.font_style: int = 64

        self.menu_name: str | None = None
        # FOND Name
        self.apple_name: str = ""
        # FOND ID
        self.fond_id: int = 128
        # OpenType-specific font Family Name
        self.pref_family_name: str | None = None
        # OpenType-specific font Style Name
        self.pref_style_name: str | None = None
        # OpenType-specific font Mac Name
        self.mac_compatible: str | None = None
        # glyph name that represents the PFM default
        self.default_character: str | None = None
        self.weight = None
        self.weight_code = -1
        self.width = "Normal"
        self.designer: str | None = None
        self.designer_url: str | None = None
        # list of font name records
        self._fontnames: ListParent[NameRecord] = ListParent(parent=self)
        # Copyright name field
        self.copyright: str | None = None
        # Notice field
        self.notice: str | None = None
        # Font note
        self.note: str | None = None
        # Type 1 Unique ID number
        self.unique_id: int = -1
        # TrueType Unique ID record
        self.tt_u_id: str | None = None
        # TrueType Version record
        self.tt_version: str | None = None
        self.trademark: str | None = None
        self._xuid = []
        self._xuid_num = 0
        # TrueType vendor code
        self._vendor: str = ""
        self.vendor_url: str | None = None
        self.version: str | None = None

        self.year: int = 0
        self.version_major: int = 1
        self.version_minor: int = 0

        self.vp_id: int = -1
        self.ms_charset: int = 0
        self.ms_id: int = 0
        # list of Panose values
        self.panose: list[int] = [0] * 10
        self.pcl_chars_set: str = " 9U"
        self.pcl_id: int = -1

        #  Dimensions
        self.upm: int = 1000
        # list of ascenders, one for each master
        self.ascender: list[int] = [750] * 16
        # list of descenders, one for each master
        self.descender: list[int] = [-250] * 16
        # list of CapHeight records, one for each master
        self.cap_height: list[int] = [700] * 16
        # list of xHeight values, one for each master
        self.x_height: list[int] = [500] * 16
        self.default_width: list[int] = [500] * 16
        self.slant_angle: float = 0.0
        self.italic_angle: float = 0.0
        self.is_fixed_pitch: int = 0
        self.underline_position: int = -100
        self.underline_thickness: int = 50

        #  Alignment
        self.blue_fuzz: list[int] = [1] * 16
        self.blue_scale: list[float] = [0.039625] * 16
        self.blue_shift: list[int] = [7] * 16

        # up until here the default values have been verified

        # number of defined blue values
        self.blue_values_num: int = 0
        # two-dimentional array of BlueValues
        # master index is top-level index
        self.blue_values: list[list[int]] = [[]]
        # number of defined OtherBlues values
        self.other_blues_num: int = 0
        # two-dimentional array of OtherBlues
        # master index is top-level index
        self.other_blues: list[list[int]] = [[]]
        # number of FamilyBlues records
        self.family_blues_num = 0
        # two-dimentional array of FamilyBlues
        # master index is top-level index
        self.family_blues: list[list[int]] = [[]]
        # number of FamilyOtherBlues records
        self.family_other_blues_num: int = 0
        # two-dimentional array of FamilyOtherBlues
        # master index is top-level index
        self.family_other_blues: list[list[int]] = [[]]
        # list of Force Bold values, one for each master
        self.force_bold: list[int] = [0]
        self.stem_snap_h_num: int = 0
        self.stem_snap_h: list[list[int]] = [[]]
        self.stem_snap_v_num: int = 0
        self.stem_snap_v: list[list[int]] = [[]]

        #  Other

        # 0 if unmodified, 1 if modified (to control the
        # 'save changes' dialog) <font color="red">(this attribute is not
        # reported by the docstring)</font>
        self.modified: int = 0
        # list of glyph classes
        self._classes = ClassList()
        # string containing the OT classes defined in the lower
        # right part of the OpenType panel <font color="red">(this description
        # is not reported by the docstring)</font>
        self.ot_classes: str = ""
        # list of OpenType features
        self._features: ListParent[Feature] = ListParent(parent=self)
        # font custom data field
        self.customdata: str = ""
        # list of custom TrueType tables
        self._truetypetables: ListParent[TrueTypeTable] = ListParent(parent=self)
        # loaded TrueType information
        # (mostly hinting-related tables)
        self.ttinfo = TTInfo()
        # current encoding of the font, read-only
        self._encoding = Encoding()
        self._encoding._parent = self
        self._encoding.load_font_default()
        # list of codepage-numbers (see OT-specs)
        # <font color="red">(this attribute is not reported by the docstring
        # this attribute is linked with the
        # ttinfo.os2_ul_code_page_range1 and
        # ttinfo.os2_ul_code_page_range2 values)</font>
        self.codepages: list[int] = []
        # list of unicoderange-numbers (see OT-specs)
        # <font color="red">(this attribute is not reported by the docstring)</font>
        self.unicoderanges: list[int] = []
        # the Created By field
        # <font color="red">(this description is not reported by the docstring)</font>
        self.source = None
        # list of MM-settings for generate instance
        self.weight_vector = WeightVector()
        self.weight_vector._parent = self
        # list of horizontal guides
        # <font color="red">(new in v4.5.4 and not reported by docstring)</font>
        self.hguides: list[Guide] = []
        # list of vertical guides
        # <font color="red">(new in v4.5.4 and not reported by docstring)</font>
        self.vguides: list[Guide] = []

        self._axis: list[tuple[str, str, str]] = []
        self._glyphs: ListParent[Glyph] = ListParent(parent=self)

        # Font data that is not accessible via FL5 Python API
        self._encoding_default: list[EncodingRecord] = []
        self._masters_count: int = 1
        self._license: str = ""
        self._license_url: str = ""
        self._custom_cmaps: list[CustomCmap] = []
        self._export_pclt_table: int = 0
        self._pclt_table: dict[str, Any] = {
            "font_number": 0,
            "pitch": 0,
            "x_height": 0,
            "style": 0,
            "type_family": 0,
            "cap_height": 0,
            "symbol_set": 0,
            "typeface": "\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000\u0000",
            "character_complement": [0] * 8,
            "file_name": "\u0000\u0000\u0000\u0000\u0000\u0000",
            "stroke_weight": 0,
            "width_type": 0,
            "serif_style": 0,
        }
        self._export_options: list[str] = []
        self._ot_export_options: list[dict[str, int]] = []
        self._axis_count: int = 0
        self._anisotropic_interpolation_mappings: list[tuple[int, int]] = []
        self._axis_mappings_count: list[int] = [0, 0, 0, 0]
        self._axis_mappings: list[tuple[float, float]] = [(0.0, 0.0)] * 40
        self._kerning_class_flags: dict[str, tuple[int, int]] = {}
        self._metrics_class_flags: dict[str, tuple[int, int, int]] = {}
        self._master_names = ["Untitled"]
        self._master_locations = [(1, (0.0, 0.0, 0.0, 0.0))]
        self._master_ps_infos = [
            {
                "font_matrix": [0.001, 0.0, 0.0, 0.001, 0.0, 0.0],
                "force_bold": 0,
                "blue_values": [0] * 14,
                "other_blues": [0] * 10,
                "family_blues": [0] * 14,
                "family_other_blues": [0] * 10,
                "blue_scale": 0.039625,
                "blue_shift": 7,
                "blue_fuzz": 1,
                "std_hw": 100,
                "std_vw": 50,
                "stem_snap_h": [0] * 12,
                "stem_snap_v": [0] * 12,
                "bounding_box": {
                    "xMin": 32767,
                    "yMin": 32767,
                    "xMax": -32767,
                    "yMax": -32767,
                },
                "adv_width_min": 0,
                "adv_width_max": 0,
                "adv_width_avg": 500,
                "ascender": 750,
                "descender": -250,
                "x_height": 500,
                "cap_height": 700,
            }
        ]

        self._mapping_mode = {
            "mapping_mode": "names_or_index",
            "2": 152,
            "3": 316,
            "mapping_id": 1,
        }
        self._primary_instance_locations: list[float] = []
        self._primary_instances: list[dict[str, Any]] = []
        self._postscript_hinting_options: dict[str, list[int] | int] = {"other": []}
        self._unknown_pleasures: dict[int, int | str | list] = {
            1502: 0,
            518: "",
            257: "",
            1140: "",
            1068: [],
            2030: "",
            513: "",
            271: "",
            527: "",
        }



from __future__ import annotations

from FL.fake.Base import Copyable


class Uni(Copyable):
    """
    Uni - class to represent Unicode index

    This class is necessary to distinguish integers and Unicode indexes. Integers are
    usually read as glyph indexes, so it is necessary to create Uni objects to access
    glyph by Unicode index

    Please note that this is not the same that is implemented in Python 2.x
    """

    __slots__ = ["_value"]

    # Constructor

    def __init__(self, uni_or_int_or_hex: Uni | int | str | None = None) -> None:
        """
        Uni() - generic constructor, creates an empty Uni record (bullshit)
        Uni(Uni) - copy constructor
        Uni(uni: int) - creates Uni object and assigns integer value
        Uni(uni_hex: string) - creates Uni object and reads value from the string in hex form
        """
        if uni_or_int_or_hex is None:
            # The error message is bullshit, but here we are:
            raise RuntimeError(
                "Incorrect # of args to: \n Uni()\n Uni(float x,float y)\n Uni(Uni P)"
            )
        if isinstance(uni_or_int_or_hex, Uni):
            self._copy_constructor(uni_or_int_or_hex)
        elif isinstance(uni_or_int_or_hex, str):
            self.value = int(uni_or_int_or_hex, base=16)
        elif isinstance(uni_or_int_or_hex, int):
            self.value = uni_or_int_or_hex

    def __repr__(self) -> str:
        return f"uni{self.value:04x}"

    # Attributes

    @property
    def value(self) -> int:
        """Unicode index"""
        return self._value

    @value.setter
    def value(self, value: int) -> None:
        self._value = value

    # Operations

    def __eq__(self, other: object) -> bool:
        """
        Compares two Unicode indexes in Uni or integer form

        Args:
            other (Uni): The object to compare to.

        Returns:
            bool: Whether both objects represent the same Unicode index.
        """
        if isinstance(other, Uni):
            return self.value == other.value

        return False

    # Methods: none

    """
    Uni type is expected in arg 1: Uni(Uni)
    Uni(Int value)
    Uni
    Uni type is expected in to compare

    class Uni has no attribute %s or it is read-only

    Incorrect type of arg assigned to value of Uni
    """



import unittest

from FL.fake.KerningClass import KerningClass


class KerningClassTests(unittest.TestCase):
    def test_instantiation_empty(self) -> None:
        kc = KerningClass()
        assert isinstance(kc, KerningClass)
        assert kc.name is None
        assert kc.sides == ""
        assert kc.keyglyph is None
        assert kc.glyphs == []

    def test_instantiation_from_flclass(self) -> None:
        kc = KerningClass(fromFLClass="_LAT_A_1st: A' Adieresis Aacute")
        assert kc.name == "_LAT_A_1st"
        assert kc.sides == "L"  # derived from "1st"
        assert kc.keyglyph == "A"
        assert kc.glyphs == ["Adieresis", "Aacute"]



import unittest

from FL.helpers.math import int32_to_unsigned, uint32_to_signed

s1 = -458254122
u1 = 3836713174

s2 = 2082841200
u2 = 2082841200


class MathTest(unittest.TestCase):
    def test_signed_to_unsigned_1(self) -> None:
        assert int32_to_unsigned(0) == 0

    def test_signed_to_unsigned_2(self) -> None:
        assert int32_to_unsigned(s1) == u1

    def test_signed_to_unsigned_3(self) -> None:
        assert int32_to_unsigned(s2) == u2

    def test_unsigned_to_signed_1(self) -> None:
        assert uint32_to_signed(0) == 0

    def test_unsigned_to_signed_2(self) -> None:
        assert uint32_to_signed(u1) == s1

    def test_unsigned_to_signed_3(self) -> None:
        assert uint32_to_signed(u2) == s2



from __future__ import absolute_import, division, print_function

from FL import fl


def getFontIndex(font):
    """
    Get the index of the supplied font.
    We must iterate through the open fonts and compare file names,
    because the == operator can not compare the font objects directly.
    (FL font objects get a different id() each time they are called)

    :param font: A title for the dialog.
    :type font:  :py:class:`FL.Font`
    """
    for i in range(len(fl)):
        cf = fl[i]
        if cf.file_name == font.file_name:
            if font.file_name is None:
                if (
                    cf.family_name == font.family_name
                    and cf.style_name == font.style_name
                ):
                    return (cf, i)
            else:
                return (cf, i)
    # Font was not found, probably there are no open fonts
    return (None, -1)


def setSelection(font, glyph_names):
    """
    Set glyphs from the glyph_names list as selected in the font window.
    """
    f, i = getFontIndex(font)
    if i > -1:
        fl.ifont = i
        fl.Unselect()
        for n in glyph_names:
            fl.Select(n)


def selectComposites(font):
    """
    Select composites in font.
    """
    setSelection(font, [glyph.name for glyph in font.glyphs if glyph.components])



import unittest

from FL.objects.TTGasp import TTGasp


class TTGaspTests(unittest.TestCase):
    def test_instantiation(self) -> None:
        # No args
        t = TTGasp()
        assert t.ppm == 1000
        assert t.behavior == 0

    def test_instantiation_copy(self) -> None:
        # Use the copy constructor
        t0 = TTGasp()
        t0.ppm = 16
        t0.behavior = 6

        t = TTGasp(t0)
        assert t.ppm == 16
        assert t.behavior == 6

    def test_instantiation_args(self) -> None:
        # Initialize with ppm and behavior
        t = TTGasp(100, 15)
        assert t.ppm == 100
        assert t.behavior == 15



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from vfbLib import replace_types
from vfbLib.enum import G
from vfbLib.typing import (
    AnchorDict,
    GdefDict,
    HintDict,
    MaskData,
    MMAnchorDict,
    MMHintsDict,
)

from FL.fake.Base import Copyable
from FL.fake.mixins import GuideMixin, GuidePropertiesMixin
from FL.helpers.ListParent import ListParent
from FL.objects.Anchor import Anchor
from FL.objects.Component import Component
from FL.objects.Hint import Hint
from FL.objects.Image import Image
from FL.objects.KerningPair import KerningPair
from FL.objects.Link import Link
from FL.objects.Node import Node
from FL.objects.Point import Point
from FL.objects.Rect import Rect
from FL.objects.Replace import Replace

if TYPE_CHECKING:
    from collections.abc import Iterable

    from vfbLib.typing import Instruction

    from FL.objects.AuditRecord import AuditRecord
    from FL.objects.Font import Font
    from FL.objects.Guide import Guide
    from FL.objects.Matrix import Matrix
    from FL.objects.TTPoint import TTPoint


logger = logging.getLogger(__name__)


class Glyph(Copyable, GuideMixin, GuidePropertiesMixin):
    """
    A glyph
    """

    __slots__ = [
        "_anchors",
        "_components",
        "_height",
        "_hguides",
        "_hhints",
        "_hlinks",
        "_index",
        "_kerning",
        "_layers_number",
        "_mask",
        "_nodes",
        "_replace_table",
        "_unknown_pleasures",
        "_vguides",
        "_vhints",
        "_vlinks",
        "_width",
        "advance_height",
        "advance_width",
        "customdata",
        "end_points",
        "flags",
        "hdmx",
        "_image",
        "_imported",
        "instructions",
        "left_side_bearing",
        "mark",
        "name",
        "note",
        "number_of_contours",
        "points",
        "top_side_bearing",
        "_unicodes",
        "y_pels",
        # Non-API
        "_glyph_bitmaps",
        "_glyph_hinting_options",
        "_glyph_origin",
        "_glyph_sketch",
        "_mask_additional",
        "_mask_metrics_mm",
        "_mask_metrics",
        "_metrics",
        "_parent",
        "_tth",
        "_write_empty_anchor_supp",
        "_write_empty_anchors_mm",
        "_write_empty_gdef",
        "_write_empty_guides",
        "_write_empty_guide_props",
        "_write_empty_links",
        "_write_empty_origin",
    ]

    # Constructor

    def __init__(
        self, glyph_or_masterscount: Glyph | int = 1, nodes: list[Node] | None = None
    ) -> None:
        self.set_defaults()

        # Process params

        if isinstance(glyph_or_masterscount, Glyph):
            self._copy_constructor(glyph_or_masterscount)

        elif isinstance(glyph_or_masterscount, int):
            self._layers_number: int = glyph_or_masterscount
            self._unknown_pleasures[2023] = [0 for _ in range(self._layers_number)]
            if nodes is not None:
                # Assign nodes
                for node in nodes:
                    self.nodes.append(node)
        # else: Empty Glyph

    def __repr__(self) -> str:
        return "<Glyph: '%s', %i nodes, orphan>" % (self.name, len(self))

    # Additions for FakeLab

    @property
    def _masters_count(self) -> int:
        return self.layers_number

    def fake_update(self, font: Font | None = None, index: int = -1) -> None:
        """
        Is called from FontLab.UpdateFont()
        """
        self._parent: Font | None = font
        self._index: int = index
        if font is None:
            self._index = -1
        for n in self.nodes:
            n.fake_update(self)

    def fake_deserialize(self, key: int, data: Any) -> None:
        """
        Add data from a VFB entry

        Args:
            name (str): The name of the entry.
            data (_type_): The entry data.
        """
        if key in (2023, 2034):
            self._unknown_pleasures[key] = data
            return

        match key:
            case 2001:
                self._layers_number = data["num_masters"]
                # 0x01
                self.name: str = data.get("name")
                # 0x02
                self._metrics = [Point(x, y) for x, y in data.get("metrics", [])]
                # 0x03
                if hint_data := data.get("hints", {}):
                    self.fake_deserialize_hints(hint_data)
                # 0x04
                if guides_data := data.get("guides", []):
                    self._write_empty_guides = True
                    self.fake_deserialize_guides(guides_data)
                # 0x05
                for component_data in data.get("components", []):
                    component = Component()
                    component.fake_deserialize(self._layers_number, component_data)
                    self.components.append(component)
                # 0x06
                for index, values in data.get("kerning", {}).items():
                    pair = KerningPair(int(index))
                    pair._values = values
                    self.kerning.append(pair)
                # 0x08
                for node_data in data["nodes"]:
                    node = Node()
                    node.fake_deserialize(self._layers_number, node_data)
                    self.nodes.append(node)
                # 0x09
                if imported := data.get("imported"):
                    self._imported = imported
                # 0x0a
                if tth := data.get("tth"):
                    self._tth = tth

            case G.Links:
                self._write_empty_links = True
                for axis, target in (("x", self.vlinks), ("y", self.hlinks)):
                    axis_links = data.get(axis, [])
                    for axis_link in axis_links:
                        target.append(Link(*axis_link))

            case G.image:
                self._image.fake_deserialize(data)
            case G.Bitmaps:
                self._glyph_bitmaps = data
            case G.Sketch:
                self._glyph_sketch = data
            case G.HintingOptions:
                self._glyph_hinting_options = data
            case G.mask:
                mask = Glyph()
                mask.fake_deserialize(G.Glyph, data)
                num = data["num"]
                self._mask_additional["num"] = num
                for i in range(num):
                    self._mask_additional[f"reserved{i}"] = data[f"reserved{i}"]
                self._mask: Glyph | None = mask
            case G.MaskMetrics:
                self._mask_metrics: Point | None = Point(*data)
            case G.MaskMetricsMM:
                self._mask_metrics_mm: list[Point] | None = [
                    Point(*coords) for coords in data
                ]
            case G.Origin:
                self._write_empty_origin = True
                self._glyph_origin = data
            case G.unicodes:
                self._unicodes.extend(data)
            case G.UnicodesNonBMP:
                self._unicodes.extend(data)
            case G.mark:
                self.mark: int = data
            case G.customdata:
                self.customdata: str | None = data
            case G.note:
                self.note: str | None = data
            case G.GDEFData:
                self._write_empty_gdef = True
                self.fake_deserialize_gdef(data)
            case G.AnchorsProperties:
                self._write_empty_anchor_supp = True
                self.fake_deserialize_anchor_supp(data)
            case G.AnchorsMM:
                self._write_empty_anchors_mm = True
                self.fake_deserialize_anchors_mm(data)
            case G.GuideProperties:
                self._write_empty_guide_props = True
                self.fake_deserialize_guide_properties(data)
            case _:
                logger.warning(f"Unhandled glyph entry: {key}")

    def fake_serialize(self) -> dict[int, Any]:  # noqa: C901
        """
        Serialize the glyph to a dict which resembles the low-level VFB structure

        Returns:
            dict[str, Any]: The serialized glyph.
        """
        # TODO: Which entries are required? Leave out the other ones.
        s: dict[int, Any] = {
            G.Glyph: {
                # Minimum wage, yeah!
                "name": self.name,
                "num_masters": self.layers_number,
                "nodes": [
                    node.fake_serialize(self.layers_number) for node in self.nodes
                ],
                "metrics": [
                    [int(p.x), int(p.y)]
                    for p in [self.GetMetrics(i) for i in range(self.layers_number)]
                ],
                "components": [comp.fake_serialize() for comp in self.components],
            },
            2023: self._unknown_pleasures[2023],
            G.HintingOptions: self._glyph_hinting_options,
        }

        # Additions for Glyph

        if self.vhints or self.hhints:
            s[G.Glyph]["hints"] = self.fake_serialize_hints()

        if self._write_empty_guides or self.hguides or self.vguides:
            s[G.Glyph]["guides"] = self.fake_serialize_guides()

        if self.kerning:
            s[G.Glyph]["kerning"] = {pair.key: pair.values for pair in self.kerning}
        imported = self._imported
        if imported:
            s[G.Glyph]["imported"] = imported

        tth = self._tth
        if tth:
            s[G.Glyph]["tth"] = tth

        # Additions related to glyph

        if self._write_empty_links or self.hlinks or self.vlinks:
            s[G.Links] = {
                "x": [[link.node1, link.node2] for link in self.vlinks],
                "y": [[link.node1, link.node2] for link in self.hlinks],
            }

        if self.image:
            image = self.image.fake_serialize()
            if image is not None:
                s[G.image] = self.image.fake_serialize()
        if self._glyph_bitmaps:
            s[G.Bitmaps] = self._glyph_bitmaps
        if self._glyph_sketch:
            s[G.Sketch] = self._glyph_sketch
        if self.mask:
            s[G.mask] = self.fake_serialize_mask()
        if self._mask_metrics:
            s[G.MaskMetrics] = (int(self._mask_metrics.x), int(self._mask_metrics.y))
        if self._mask_metrics_mm:
            s[G.MaskMetricsMM] = [(int(p.x), int(p.y)) for p in self._mask_metrics_mm]

        origin = self._glyph_origin
        if self._write_empty_origin or origin != {"x": 0, "y": 0}:
            s[G.Origin] = origin

        unicodes = [u for u in self.unicodes if u <= 0xFFFF]
        if unicodes:
            s[G.unicodes] = unicodes

        if 2034 in self._unknown_pleasures:
            s[2034] = self._unknown_pleasures[2034]

        unicodes_non_bmp = [u for u in self.unicodes if u > 0xFFFF]
        if unicodes_non_bmp:
            s[G.UnicodesNonBMP] = unicodes_non_bmp

        if self.mark != 0:
            s[G.mark] = self.mark

        if self.customdata:
            s[G.customdata] = self.customdata

        if self.note:
            s[G.note] = self.note

        gdef = self.fake_serialize_gdef()
        if self._write_empty_gdef or (
            gdef["anchors"]
            or gdef["carets"]
            or gdef["unknown"]
            or gdef["glyph_class"] not in (None, "unassigned")
        ):
            s[G.GDEFData] = gdef

        anchors_supp = self.fake_serialize_anchor_supp()
        if self._write_empty_anchor_supp or anchors_supp:
            s[G.AnchorsProperties] = anchors_supp

        if self.anchors and self.layers_number > 1 or self._write_empty_anchors_mm:
            s[G.AnchorsMM] = self.fake_serialize_anchors_mm()

        guide_properties = self.fake_serialize_guide_properties()
        # Only write if there are glyph guides
        if (
            self._write_empty_guides
            or self.hguides
            or self.vguides
            and (
                self._write_empty_guide_props
                or guide_properties["h"]
                or guide_properties["v"]
            )
        ):
            s[G.GuideProperties] = guide_properties

        return s

    def fake_deserialize_hints(self, data: MMHintsDict) -> None:
        for direction, target in (("v", self.vhints), ("h", self.hhints)):
            for hint_list in data[direction]:
                hint = Hint()
                hint.fake_deserialize(hint_list)
                target.append(hint)
        for record_type, index in data.get("hintmasks", []):
            self.replace_table.append(
                Replace({"h": 0x01, "v": 0x02, "r": 0xFF}[record_type], index)
            )

    def fake_serialize_hints(self) -> MMHintsDict:
        hints_dict = MMHintsDict(v=[], h=[])
        for direction, source in (("v", self.vhints), ("h", self.hhints)):
            for hint in source:
                mmhint = []
                for master_index in range(self._masters_count):
                    mmhint.append(
                        HintDict(
                            pos=hint._positions[master_index],
                            width=hint._widths[master_index],
                        )
                    )
                hints_dict[direction].append(mmhint)
        if self.replace_table:
            hintmasks = hints_dict["hintmasks"] = []
            for r in self.replace_table:
                hintmasks.append((replace_types[r.type], r.index))
        return hints_dict

    def fake_serialize_mask(self) -> MaskData:
        num = self._mask_additional["num"]
        mask = MaskData(num_masters=self.layers_number, num=num)
        for i in range(num):
            mask[f"reserved{i}"] = self._mask_additional[f"reserved{i}"]
        if self.mask is None:
            return mask

        mask["nodes"] = [
            node.fake_serialize(self.layers_number) for node in self.mask.nodes
        ]
        return mask

    def fake_deserialize_gdef(self, data: GdefDict) -> None:
        for anchor_dict in data.get("anchors", []):
            self.anchors.append(
                Anchor(anchor_dict.get("name", ""), anchor_dict["x"], anchor_dict["y"])
            )
        for caret in data.get("carets", []):
            self._carets.append(caret)
        glyph_class = data.get("glyph_class")
        if glyph_class:
            self._gdef_class: str | None = glyph_class
        self._gdef_unknown.extend(data.get("unknown", []))

    def fake_serialize_gdef(self) -> GdefDict:
        gdef = GdefDict(anchors=[], carets=[], glyph_class="unassigned", unknown=[])
        for anchor in self.anchors:
            gdef["anchors"].append(
                AnchorDict(name=anchor.name, x=anchor.x, x1=-1, y=anchor.y, y1=-1)
            )
        gdef["carets"] = self._carets
        gdef["glyph_class"] = self._gdef_class
        gdef["unknown"].extend(self._gdef_unknown)
        return gdef

    def fake_deserialize_anchor_supp(self, data: list[dict[str, int]]) -> None:
        for i, d in enumerate(data):
            anchor = self.anchors[i]
            anchor.mark = d["hue"]
            anchor._reserved = d["reserved"]

    def fake_serialize_anchor_supp(self) -> list[dict[str, int]]:
        data = []
        for anchor in self.anchors:
            data.append({"hue": anchor.mark, "reserved": anchor._reserved})
        return data

    def fake_deserialize_anchors_mm(self, data: MMAnchorDict) -> None:
        for i, anchor in enumerate(data):
            target = self.anchors[i]._points
            ax = anchor["x"]
            ay = anchor["y"]
            for master_index in range(self._masters_count):
                target[master_index].x = ax[master_index]
                target[master_index].y = ay[master_index]

    def fake_serialize_anchors_mm(self) -> list[MMAnchorDict]:
        anchors = []
        for anchor in self.anchors:
            anchor_dict = MMAnchorDict(x=[], y=[])
            for p in anchor._points:
                anchor_dict["x"].append(int(p.x))
                anchor_dict["y"].append(int(p.y))
            anchors.append(anchor_dict)
        return anchors

    # Attributes

    @property
    def parent(self) -> Font | None:
        """
        The glyph's parent object.

        Returns:
            Font | None: The parent :py:class:`Font` or None.
        """
        return self._parent

    @property
    def nodes(self) -> ListParent[Node]:
        """
        The list of the glyph's nodes.

        Returns:
            ListParent[Node]: The nodes.
        """
        return self._nodes

    # customdata
    # note
    # mark

    @property
    def anchors(self) -> list[Anchor]:
        """
        The list of the glyph's anchos.

        Returns:
            list[Anchor]: The anchors.
        """
        return self._anchors

    @property
    def hhints(self) -> ListParent[Hint]:
        """
        The list of the glyph's horizontal stem hints.

        Returns:
            ListParent[Hint]: The list of hints.
        """
        return self._hhints

    @property
    def vhints(self) -> ListParent[Hint]:
        """
        The list of the glyph's vertical stem hints.

        Returns:
            ListParent[Hint]: The list of hints.
        """
        return self._vhints

    @property
    def hlinks(self) -> ListParent[Link]:
        """
        The list of the glyph's horizontal stem links.

        Returns:
            ListParent[Link]: The list of links.
        """
        return self._hlinks

    @property
    def vlinks(self) -> ListParent[Link]:
        """
        The list of the glyph's vertical stem links.

        Returns:
            ListParent[Link]: The list of links.
        """
        return self._vlinks

    @property
    def hguides(self) -> ListParent[Guide]:
        """
        The list of the glyph's horizontal guides.

        Returns:
            ListParent[Guide]: The list of guides.
        """
        return self._hguides

    @property
    def vguides(self) -> ListParent[Guide]:
        """
        The list of the glyph's vertical guides.

        Returns:
            ListParent[Guide]: The list of guides.
        """
        return self._vguides

    @property
    def components(self) -> ListParent[Component]:
        """
        The list of the glyph's components.

        Returns:
            ListParent[Component]: The list of components.
        """
        return self._components

    @property
    def replace_table(self) -> list[Replace]:
        """
        The hint replacement program, a list of :py:class:`Replace` objects.

        Returns:
            list[Replace]: The list of replace objects.
        """
        return self._replace_table

    @property
    def kerning(self) -> ListParent[KerningPair]:
        """
        The list of the glyph's kerning pairs.

        Returns:
            ListParent[KerningPair]: The list of kerning pairs.
        """
        return self._kerning

    @property
    def layers_number(self) -> int:
        """
        The number of masters for the glyph.

        Returns:
            int: The number of masters.
        """
        return self._layers_number

    @property
    def mask(self) -> Glyph | None:
        """
        The mask (background layer) of the glyph.

        Returns:
            Glyph | None: The mask glyph if present.
        """
        return self._mask

    # flags

    @property
    def nodes_number(self) -> int:
        """
        The number of nodes in the glyph, same as 'len(Glyph)'.

        Returns:
            int: The number of nodes.
        """
        return len(self._nodes)

    @property
    def width(self) -> int:
        """
        The advance width for the first master.

        Returns:
            int: _description_
        """
        return int(self.GetMetrics().x)

    @width.setter
    def width(self, value: int) -> None:
        # TODO: Does this set the first master only?
        self._width = value

    @property
    def height(self) -> int:
        """
        advance height for the first master

        Returns:
            int: _description_
        """
        return int(self.GetMetrics().y)

    @height.setter
    def height(self, value: int) -> None:
        # TODO: Does this set the first master only?
        self._height = value

    @property
    def unicode(self) -> int | None:
        """
        The first Unicode index as integer.

        Returns:
            int | None: The Unicode codepoint.
        """
        if self._unicodes:
            return self._unicodes[0]
        else:
            return None

    @unicode.setter
    def unicode(self, value: int) -> None:
        if not isinstance(value, int):
            raise TypeError
        if value == -1:
            return
        if not self._unicodes:
            self._unicodes.append(value)
        else:
            self._unicodes[0] = value

    @property
    def unicodes(self) -> list[int]:
        return self._unicodes

    @unicodes.setter
    def unicodes(self, value: list[int]) -> None:
        self._unicodes = value.copy()

    # name

    @property
    def image(self) -> Image:
        return self._image

    @image.setter
    def image(self, value: Image) -> None:
        if not isinstance(value, Image):
            raise TypeError

        self._image = value

    @property
    def index(self) -> int:
        """
        The glyph index in the font.

        Returns:
            int: The glyph index or -1 if the glyph has no parent font.
        """
        if self.parent is None:
            return -1
        return self._index

    # TrueType Data:

    # advance_width
    # advance_height
    # left_side_bearing
    # top_side_bearing
    # y_pels

    @property
    def bounding_box(self) -> Rect:
        return self.GetBoundingRect(0)

    # number_of_contours
    # end_points
    # points
    # instructions
    # hdmx

    # Operations

    def __len__(self) -> int:
        """
        The number of nodes in the glyph.

        Returns:
            int: The number of nodes
        """
        return len(self._nodes)

    def __getitem__(self, index: int) -> Node:
        """
        Access the nodes list

        Args:
            index (int): The node index.

        Returns:
            Node: The node at index.
        """
        return self._nodes[index]

    def __slice__(self, a: int, b: int) -> list[Node]:
        """
        Access a slice of the glyph's nodes list.

        Args:
            a (int): The first node index.
            b (int): The last node index.

        Returns:
            list[Node]: The slice of nodes.
        """
        raise NotImplementedError

    def __add__(self, glyph_node_nodelist: Glyph | Node | Iterable[Node]) -> Glyph:
        """
        Append a :py:class:`Glyph`, a :py:class:`Node`, or a sequence of
        :py:class:`Node`s.

        Args:
            glyph_node_nodelist (Glyph | Node | Iterable[Node]): The objects to append.

        Returns:
            Glyph: The resulting glyph.
        """
        raise NotImplementedError

    def __mul__(self, matrix: Matrix) -> Glyph:
        """
        Apply a matrix transformation to the glyph.

        Args:
            matrix (Matrix): The transformation to apply.

        Returns:
            Glyph: The transformed glyph.
        """
        raise NotImplementedError

    # Methods

    def Assign(self, g: Glyph) -> None:
        """
        Copy all information from a glyph.

        Args:
            g (Glyph): The source glyph.
        """
        raise NotImplementedError

    def Transform(self, matrix: Matrix) -> None:
        """
        Apply a matrix transformation to the glyph.

        Args:
            matrix (Matrix): The transformation to apply.
        """
        raise NotImplementedError

    def SetLayersNumber(self, mastersnumber: int) -> None:
        """
        Change the number of masters. Only applicable to glyphs that have no parent.

        Args:
            mastersnumber (int): The new number of masters.
        """
        raise NotImplementedError

    def Clear(self) -> None:
        """
        Remove all nodes.
        """
        raise NotImplementedError

    def Add(self, glyph_node_nodelist: Glyph | Node | Iterable[Node]) -> None:
        """
        Append a glyph, a node, or a sequence of nodes.

        Args:
            glyph_node_nodelist (Glyph | Node | Iterable[Node]): The objects to append.
        """
        raise NotImplementedError

    def Insert(
        self, node_or_glyph_or_nodelist: Node | Glyph | list[Node], nodeindex: int = 0
    ) -> None:
        """
        Insert a node, a glyph, or a sequence of nodes at the beginning of the glyph's
        nodes, or at the specified node index.

        Args:
            node_or_glyph_or_nodelist (Node | Glyph | list[Node]): The nodes to insert.
            nodeindex (int, optional): The place to insert. Defaults to 0.
        """
        raise NotImplementedError

    def Present(self, style) -> bool:
        """
        Return True if a layer or a combination of layers are present in the glyph.

        Args:
            style (_type_): _description_

        Returns:
            bool: Whether the specified combination of layers is present.
        """
        raise NotImplementedError

    def Delete(self, index0: int, index1: int | None = None) -> None:
        """
        Remove a node or a range of nodes.

        Args:
            index0 (int): The first node index
            index1 (int | None, optional): The last node index. Defaults to None.
        """
        raise NotImplementedError

    def ExpandLayer(self, masterindex: int) -> None:
        """
        Expand the specified master to all other masters.

        Args:
            masterindex (int): The master index.
        """
        raise NotImplementedError

    def Shift(self, point: Point, masterindex: int = 0) -> None:
        """
        Shift the positions of all nodes in the specified master.

        Args:
            point (Point): The point specifying the x and y shift.
            masterindex (int, optional): The master to shift. Defaults to 0.
        """
        raise NotImplementedError

    def Scale(self, scale: Point, center: Point, masterindex: int = 0) -> None:
        """
        Scale the glyph.

        Args:
            scale (Point): The point specifying the x and y scale factor.
            center (Point): The center of the scale transformation.
            masterindex (int, optional): The master to scale. Defaults to 0.
        """
        raise NotImplementedError

    def Layer(self, masterindex: int) -> list[Point]:
        """
        Return a list of :py:class:`Points` for all nodes for the selected master.

        Args:
            masterindex (int): The master index

        Returns:
            list[Point]: The list of each node's last point.
        """
        raise NotImplementedError

    def Section(self, masterindex: int, pointindex: int, nodetype: int) -> list[Point]:
        """
        Return a list of points matching the specified options.

        Args:
            masterindex (int): The master index.
            pointindex (int): The point index.
            nodetype (int): The node type (nLINE, nMOVE, nCURVE, nOFF).

        Returns:
            list[Point]: The list of points.
        """
        raise NotImplementedError

    def MoveNode(self, options) -> None:
        """
        Moves the node copying Edit tool behavior (see User manual for details).
        """
        raise NotImplementedError

    def DeleteNode(self, nodeindex: int) -> None:
        """
        Remove the specified Node.

        Args:
            nodeindex (int): The node index.
        """
        raise NotImplementedError

    def InsertNode(
        self, nodeindex: int, time: float = 0.0, masterindex: int = 0
    ) -> None:
        """
        Insert a new node on a contour.

        Args:
            nodeindex (int): _description_
            time (float, optional): The time on the segment. Start is 0.0, end is 1.0.
                Defaults to 0.0.
            masterindex (int, optional): The master index used for measuring. Defaults
                to 0.
        """
        raise NotImplementedError

    # SELECTION-METHODS

    def Selection(self) -> list[Node]:
        """
        Return a list of selected nodes.

        Returns:
            list[Node]: The list of nodes.
        """
        raise NotImplementedError

    def SelectAll(self) -> None:
        """
        Select all nodes.
        """
        raise NotImplementedError

    def UnselectAll(self) -> None:
        """
        Deselect all nodes.
        """
        raise NotImplementedError

    def InvertSelection(self) -> None:
        """
        Select unselected nodes and deselect selected nodes.
        """
        raise NotImplementedError

    def isAnySelected(self) -> bool:
        """
        Return True if at least one node is selected.
        """
        raise NotImplementedError

    def SelectedCount(self) -> int:
        """
        Return the number of selected nodes.
        """
        raise NotImplementedError

    def SelectRect(self, r: Rect, masterindex: int = 0) -> None:
        """
        Select all nodes that are inside a rectangle.

        Args:
            r (Rect): The rectangle.
            masterindex (int, optional): The master index used for measuring. Defaults
                to 0.
        """
        raise NotImplementedError

    def UnselectRect(self, r: Rect, masterindex: int = 0) -> None:
        """
        Deselect all nodes that are inside a rectangle.

        Args:
            r (Rect): The rectangle.
            masterindex (int, optional): The master index used for measuring. Defaults
                to 0.
        """
        raise NotImplementedError

    def DeleteSelected(self) -> None:
        """
        Delete all selected nodes.
        """
        raise NotImplementedError

    # METRICS-METHODS

    def GetBoundingRect(self, masterindex: int = 0) -> Rect:
        """
        Return the glyph's bounding box.

        Args:
            masterindex (int, optional): The master index used for measuring. Defaults
                to 0.

        Returns:
            Rect: The bounding box of the glyph.
        """
        if masterindex != 0:
            raise NotImplementedError

        rect = Rect(32767, 32767, -32767, -32767)
        for n in self._nodes:
            for p in n.points:
                rect += p

        return rect

    def GetMetrics(self, masterindex: int = 0) -> Point:
        """
        Return the glyph's advance width and advance height as a point.

        Args:
            masterindex (int, optional): The master index used for measuring. Defaults
                to 0.

        Returns:
            Point: The point containing the advance width (x) and heigh (y).
        """
        return self._metrics[masterindex]

    def SetMetrics(self, p: Point, masterindex: int = 0) -> None:
        """
        Assign new values to the advance width and advance height of the glyph.

        Args:
            p (Point): The point containing the advance width (x) and heigh (y).
            masterindex (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    def GetVSB(self, masterindex: int = 0) -> int:  # TODO: Is it int?
        """
        Return the glyph' bottom sidebearing position.

        Args:
            masterindex (int, optional): The master index used for measuring. Defaults
                to 0.

        Returns:
            int: The bottom sidebearing.
        """
        raise NotImplementedError

    def SetVSB(self, value: int, masterindex: int = 0) -> None:  # TODO: Is it int?
        """
        Assign new values to the bottom sidebearing of the glyph.

        Args:
            value (int): The bottom sidebearing.
            masterindex (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    # OVERLAP-METHODS

    def RemoveOverlap(self, masterindex: int = 0) -> None:
        """
        Remove overlaps from the glyph's contours.

        Args:
            masterindex (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    def Badd(self, glyph_nodelist: Glyph | list[Node], masterindex: int = 0) -> None:
        """
        Perform a boolean _Add_ operation with the glyph or list of nodes.

        Args:
            glyph_nodelist (Glyph | list[Node]): _description_
            masterindex (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    def Bsubtract(
        self, glyph_nodelist: Glyph | list[Node], masterindex: int = 0
    ) -> None:
        """
        Perform boolean _Subtract_ operation with the glyph or list of nodes.

        Args:
            glyph_nodelist (Glyph | list[Node]): _description_
            masterindex (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    def Bintersect(
        self, glyph_nodelist: Glyph | list[Node], masterindex: int = 0
    ) -> None:
        """
        Perform a boolean _Intersect_ operation with the glyph or list of nodes.

        Args:
            glyph_nodelist (Glyph | list[Node]): _description_
            masterindex (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    # CONTOUR-METHODS

    def GetContoursNumber(self) -> int:
        """
        Return the number of contours in the glyph.
        """
        raise NotImplementedError

    def GetContourBegin(self, contourindex: int) -> int:
        """
        Return the index of the first node of the speficied contour.

        Args:
            contourindex (int): The contour index.

        Returns:
            int: The node index.
        """
        raise NotImplementedError

    def GetContourLength(self, contourindex: int) -> int:
        """
        Return the number of nodes in the specified contour.

        Args:
            contourindex (int): The contour index.

        Returns:
            int: The number of nodes.
        """
        raise NotImplementedError

    def SelectContour(self, contourindex: int) -> None:
        """
        Select all nodes in the specified contour.

        Args:
            contourindex (int): The contour index.
        """
        raise NotImplementedError

    def DeleteContour(self, contourindex: int) -> None:
        """
        Remove in the specified contour.

        Args:
            contourindex (int): The contour index.
        """
        raise NotImplementedError

    def ReverseContour(self, contourindex: int) -> None:
        """
        Reverse the specified contour's path direction.

        Args:
            contourindex (int): The contour index.
        """
        raise NotImplementedError

    def ReorderContour(self, contourindex: int, newindex: int) -> None:
        """
        Reorder the contours in the glyph by moving the specified contour to a new
        index.

        Args:
            contourindex (int): The contour index.
            newindex (int): The new contour index.
        """
        raise NotImplementedError

    def isContourClockwise(self, contourindex: int) -> bool:
        """
        Return True if direction of contour is clockwise.

        Args:
            contourindex (int): The contour index.

        Returns:
            bool: True if the path direction is clockwise.
        """
        raise NotImplementedError

    def SetStartNode(self, nodeindex: int) -> None:
        """
        Make the specified node the starting node of its contour.

        Args:
            nodeindex (int): The node index.
        """
        raise NotImplementedError

    def FindContour(self, nodeindex: int) -> int:
        """
        Return number of contour containing the speficied node.

        Args:
            nodeindex (int): The node index

        Returns:
            int: The contour index.
        """
        raise NotImplementedError

    # HINTS-METHODS

    def RemoveHints(self, mode: int) -> None:
        """
        Remove hints and links.

        Args:
            mode (int): _description_
        """
        raise NotImplementedError

    def Autohint(self, masterindex: int = 0) -> None:
        """
        Automatically generate PostScript hints for the glyph.

        Args:
            masterindex (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    # ANCHOR-METHODS

    def FindAnchor(self, name: str) -> Anchor:  # XXX: does it return an anchor?
        """
        Find an anchor by name.

        Args:
            name (str): The anchor name

        Returns:
            Anchor: _description_
        """
        raise NotImplementedError

    # TRANSFORMATION-METHODS

    def Decompose(self) -> None:
        """
        Paste all components to the glyph outline.
        """
        raise NotImplementedError

    def MakeExtremeNodes(self, masterindex: int = 0) -> None:
        """
        Automatically add nodes at contour extrema.

        Args:
            masterindex (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    def Audit(self) -> list[AuditRecord]:
        """
        Perform a test of the glyph and return a list of :py:class:`AuditRecord`
        objects.

        Returns:
            list[AuditRecord]: The list of potential contour problems.
        """
        raise NotImplementedError

    def Iterate(self, iterator: Any) -> None:
        """
        Iterate the glyph through an iterator class which must provide the following
        methods:

        - :py:func:`Start()`
        - :py:func:`ClosePath()`
        - :py:func:`StartPath(Node)`
        - :py:func:`LineTo(Node)`
        - :py:func:`CurveTo(Node)`
        - :py:func:`SplineTo(Node)`
        - :py:func:`Finish()`

        Args:
            iterator (Any): The iterator object.
        """
        raise NotImplementedError

    def Rotate3D(self) -> None:
        """
        _summary_
        """
        raise NotImplementedError

    def Extrude3D(self, outlinewidth: int, shift_x: int, shift_y: int) -> None:
        """
        _summary_

        Args:
            outlinewidth (int): _description_
            shift_x (int): _description_
            shift_y (int): _description_
        """
        raise NotImplementedError

    def Shadow(self, outlinewidth: int, shift_x: int, shift_y: int) -> None:
        """
        _summary_

        Args:
            outlinewidth (int): _description_
            shift_x (int): _description_
            shift_y (int): _description_
        """
        raise NotImplementedError

    def College(self, outlinewidth: int, distance: int) -> None:
        """
        _summary_

        Args:
            outlinewidth (int): _description_
            distance (int): _description_
        """
        raise NotImplementedError

    def Gradient(
        self,
        outlinewidth: int,
        direction: int,
        stripes_number: int,
        start_y: int,
        finish_y: int,
    ) -> None:
        """
        _summary_

        Args:
            outlinewidth (int): _description_
            direction (int): _description_
            stripes_number (int): _description_
            start_y (int): _description_
            finish_y (int): _description_
        """
        raise NotImplementedError

    def Distance(
        self, width_x: int, width_y: int, cornermode: int, dest: Glyph | None = None
    ) -> None:
        """
        _summary_

        Args:
            width_x (int): _description_
            width_y (int): _description_
            cornermode (int): _description_
            dest (Glyph | None, optional): _description_. Defaults to None.
        """
        raise NotImplementedError

    def Interpolate(self, nodes: list[tuple[int, Point]]) -> list[tuple[int, Point]]:
        """
        __summary__

        Args:
            nodes (list[tuple[int, Point]]): _description_

        Returns:
            list[tuple[int, Point]]: _description_
        """
        raise NotImplementedError

    def Warp(self, points: list[Point], force: float) -> None:
        """
        _summary_

        Args:
            points (list[Point]): _description_
            force (float): _description_
        """
        raise NotImplementedError

    def Rasterize(self, image: Image) -> None:
        """
        _summary_

        Args:
            image (Image): _description_
        """
        raise NotImplementedError

    def Blend(self, source: Glyph, layer1: int, layer2: int, amount: Point) -> Glyph:
        """
        Return a blend of the glyph and source.

        Args:
            source (Glyph): _description_
            layer1 (int): _description_
            layer2 (int): _description_
            amount (Point): _description_

        Returns:
            Glyph: The blended glyph.
        """
        raise NotImplementedError

    def JoinAll(self) -> None:
        """
        Tries to join all open contours.
        """
        raise NotImplementedError

    def SaveEPS(self, filename: str, layer: int = 0) -> None:
        """
        Write a glyph master into the EPS file named filename.

        Args:
            filename (str): The path and file name.
            layer (int, optional): The master index. Defaults to 0.
        """
        raise NotImplementedError

    def LoadEPS(self, filename: str) -> Glyph:
        """
        Read an EPS file from filename and return it as a Glyph object.

        Use the :py:meth:`Assign` method to replace current the glyph with the
        imported outline.

        Args:
            filename (str): The path and file name.

        Returns:
            Glyph: The imported glyph.
        """
        raise NotImplementedError

    def R(self, points: list[Point], force: float = 0.0) -> None:
        """
        Present in the docstring but raises an :py:class:`AttributeError`.

        Args:
            points (list[Point]): A list of points.
            force (float, optional): The force. Defaults to 0.0.

        Raises:
            AttributeError: Always.
        """
        raise AttributeError

    # Additional methods reported by dir(fl.glyph)

    def EditMask(self) -> None:
        raise NotImplementedError

    def ExchangeMask(self) -> None:
        raise NotImplementedError

    def clear(self) -> None:
        raise NotImplementedError

    # FakeLab Defaults

    def set_defaults(self) -> None:
        self._parent = None
        self._nodes: ListParent[Node] = ListParent([], self)

        # custom data defined for this glyph
        self.customdata = None

        # note defined for this glyph
        self.note = None

        self.mark = 0
        self._anchors: list[Anchor] = []
        self._hhints: ListParent[Hint] = ListParent([], self)
        self._vhints: ListParent[Hint] = ListParent([], self)
        self._hlinks: ListParent[Link] = ListParent([], self)
        self._vlinks: ListParent[Link] = ListParent([], self)
        self._hguides: ListParent[Guide] = ListParent([], self)
        self._vguides: ListParent[Guide] = ListParent([], self)
        self._components: ListParent[Component] = ListParent([], self)
        self._replace_table: list[Replace] = []
        self._kerning: ListParent[KerningPair] = ListParent([], self)
        self._layers_number = 1
        self._mask = None
        self._mask_additional: dict[str, int] = {}
        self._mask_metrics = None
        self._mask_metrics_mm = None

        # flags set for this glyph
        self.flags: int = 0

        # list of Unicode indexes
        self._unicodes: list[int] = []

        # glyph name
        self.name = ""

        # [Image]           - background image (new in FL 4.53 Win)
        self._image: Image = Image(24, 24)

        # glyph index, -1 if orphan glyph (not reported by docstring)
        self._index = -1

        # TrueType data
        # probably all are only present for imported TTFs

        self.advance_width: int = 0
        self.advance_height: int = 0
        self.left_side_bearing: int = 0
        self.top_side_bearing: int = 0
        self.y_pels: int = 1
        self.number_of_contours: int = 0
        self.end_points: list[int] = []
        self.points: list[TTPoint] = []
        self.instructions: list[int] = []
        self._imported: dict | None = None
        self.hdmx: list[int] = []

        self._carets: list[tuple[int, int]] = []
        self._gdef_class = None
        self._gdef_unknown: list[int] = []
        self._glyph_bitmaps = None
        self._glyph_hinting_options = {}
        self._glyph_origin = {"x": 0, "y": 0}
        self._glyph_sketch = None
        self._tth: list[Instruction] = []

        self._unknown_pleasures: dict[int, Any] = {
            2023: [0 for _ in range(self.layers_number)]
        }

        # For binary compatibility with FL-written files:

        self._write_empty_anchors_mm = False
        self._write_empty_anchor_supp = False
        self._write_empty_gdef = False
        self._write_empty_guides = False
        self._write_empty_guide_props = False
        self._write_empty_links = False
        self._write_empty_origin = False



from __future__ import annotations

from typing import TYPE_CHECKING

from vfbLib.typing import HintDict

from FL.fake.Base import Copyable
from FL.objects.Link import Link

if TYPE_CHECKING:
    from FL.objects.Glyph import Glyph
    from FL.objects.Matrix import Matrix


class Hint(Copyable):
    """
    Hint - class to represent hint

    This class is Multiple Master - compatible
    """

    __slots__ = ["_parent", "_positions", "_widths"]

    # Constructor

    def __init__(
        self, hint_or_position: Hint | int | None = None, width: int | None = None
    ) -> None:
        """
        Hint()                - generic constructor,
                                creates a Hint with zero coordinates
        Hint(Hint)            - copy constructor
        Hint(position, width) - creates a Hint and assigns position and width values

        Args:
            hint_or_position (Hint | int | None): _description_
            width (int | None, optional): _description_. Defaults to None.
        """
        self._parent: Glyph | None = None
        self._positions: list[int] = [0] * 16
        self._widths: list[int] = [21] * 16

        arg1 = hint_or_position

        # FIXME:
        # If the second argument is present, it determines what is expected as the
        # first argument:
        # h = Hint()
        # b = Hint(h, 1)
        # TypeError: int() argument must be a string or a number, not 'Hint'
        #
        # We have it backwards at the moment: If the first argument is a Hint, the
        # second argument is ignored.

        if arg1 is not None:
            if isinstance(arg1, Hint):
                self._copy_constructor(arg1)
            elif isinstance(arg1, int):
                self.position = arg1
                if width is not None:
                    self.width = width
            else:
                raise RuntimeError("Hint type is expected in arg 1: Hint(Hint)")

    def __repr__(self) -> str:
        if self._parent is None:
            return f"<Hint: p={self.position}, w={self.width}, orphan>"
        else:
            # TODO: return HHint or VHint
            return (
                f"<Hint: p={self.position}, w={self.width}, "
                f'parent: "{self._parent.name}">'
            )

    # Additions for FakeLab

    def fake_deserialize(self, data: list[HintDict]) -> None:
        self._positions = []
        self._widths = []
        for hint_dict in data:
            self._positions.append(hint_dict["pos"])
            self._widths.append(hint_dict["width"])

    def fake_serialize(self) -> list[HintDict]:
        hint_dicts = []
        hint_dict = HintDict()
        hint_dicts.append(hint_dict)
        return hint_dicts

    # Attributes

    @property
    def parent(self) -> Glyph | None:
        """
        Hint's parent object, `Glyph`

        Returns:
            Glyph | None: _description_
        """
        return self._parent

    @property
    def position(self) -> int:
        """
        Position of the hint

        Returns:
            int: The position of the hint in the first master
        """
        return self._positions[0]

    @position.setter
    def position(self, value: int) -> None:
        # Sets the position for all masters
        # TODO: Must we keep the list, or could we just replace it?
        for i in range(16):
            self._positions[i] = value

    @property
    def width(self) -> int:
        """
        Width of the hint

        Returns:
            int: The width of the hint in the first master
        """
        return self._widths[0]

    @width.setter
    def width(self, value: int) -> None:
        # Sets the width for all masters
        # TODO: Must we keep the list, or could we just replace it?
        for i in range(16):
            self._widths[i] = value

    @property
    def positions(self) -> list[int]:
        """
        List of positions for each master

        Returns:
            list[int]: _description_
        """
        return self._positions

    @positions.setter
    def positions(self, value: list[int]) -> None:
        raise RuntimeError(
            'Attempt to write read only attribute "positions" of class Hint'
        )

    @property
    def widths(self) -> list[int]:
        """
        List of widths for each master

        Returns:
            list[int]: _description_
        """
        return self._widths

    @widths.setter
    def widths(self, value: list[int]) -> None:
        raise RuntimeError(
            'Attempt to write read only attribute "widths" of class Hint'
        )

    # Operations: none

    # Methods

    def ToLink(self) -> Link:
        """
        Transforms hint to Link (and returns it as a result) using parent as a source of
        node coordinates. Parent must exist

        Returns:
            Link: The Link.
        """
        # TODO: If this works at all in FontLab, I haven't managed to get it working
        raise RuntimeError(
            "In order to be converted to a Link, Hint must not be an orphan"
        )

    def Transform(self, m: Matrix) -> None:
        """
        Apply Matrix transformation to the Hint (see Matrix().__doc__)

        Args:
            m (Matrix): _description_
        """
        raise NotImplementedError

    def TransformLayer(self, m: Matrix, layernum: int) -> None:
        """
        Apply Matrix transformation to the selected layer of the Hint

        Args:
            m (Matrix): _description_
            layernum (int): _description_
        """
        raise NotImplementedError
